"""
Main top level calculation functions for light propagation through optical data.


"""
from __future__ import absolute_import, print_function, division
import time
from dtmm.conf import DTMMConfig,  BETAMAX, SMOOTH, FDTYPE, get_default_config_option
from dtmm.wave import k0
from dtmm.data import uniaxial_order, refind2eps, validate_optical_data
from dtmm.tmm import E2H_mat, projection_mat, alphaf
from dtmm.tmm3d import transfer3d
from dtmm.linalg import  dotmf, dotmv
from dtmm.print_tools import print_progress
from dtmm.diffract import diffract, projection_matrix, diffraction_alphaffi
from dtmm.field import field2intensity, field2betaphi, field2fvec
from dtmm.fft import fft2, ifft2
from dtmm.jones import jonesvec, polarizer
from dtmm.jones4 import ray_jonesmat4x4
from dtmm.data import effective_data
import numpy as np
from dtmm.denoise import denoise_fftfield, denoise_field

from dtmm.propagate_4x4 import propagate_4x4_full, propagate_4x4_effective_1,\
    propagate_4x4_effective_2,propagate_4x4_effective_3,propagate_4x4_effective_4
from dtmm.propagate_2x2 import propagate_2x2_full, propagate_2x2_effective_1,propagate_2x2_effective_2

#norm flags
DTMM_NORM_FFT = 1<<0 #normalize in fft mode
DTMM_NORM_REF = 1<<1 #normalize using reference field

def _isotropic_effective_data(data):
    d, material, angles = data
    n = len(d)
    epseff = uniaxial_order(0.,material).mean(axis = (0,1,2))
    epseff = np.broadcast_to(epseff,(n,3)).copy()#better to make copy.. to make it c contiguous
    aeff = np.array((0.,0.,0.))
    aeff = np.broadcast_to(aeff,(n,3)).copy()#better to make copy.. to make it c contiguous
    return validate_optical_data((d,epseff,aeff), homogeneous = True)

def _validate_betaphi(beta,phi, extendeddim = 0):
    if beta is None or phi is None:
        raise ValueError("Both beta and phi must be defined!")
    beta = np.asarray(beta, FDTYPE)
    phi = np.asarray(phi, FDTYPE)  
    
    if beta.ndim != phi.ndim:
        raise ValueError("Beta nad phi should have same dimensions!")
    
    if beta.ndim == 1:
        if len(beta) != len(phi):
            raise ValueError("Beta nad phi should have same length!")
        #make arrays broadcastable to field by adding extra dimensions
        for i in range(extendeddim):
            beta = beta[...,None]
            phi = phi[...,None]
    elif beta.ndim != 0:
        raise ValueError("Only length 1 or scalar values are supported for beta and phi")
    else:
        for i in range(extendeddim+1):
            beta = beta[...,None]
            phi = phi[...,None]        
    return beta, phi

def normalize_field(field, intensity_in, intensity_out, out = None):
    m = intensity_out == 0.
    intensity_out[m] = 1.
    intensity_in[m] = 0.
    fact = (intensity_in/intensity_out)[...,None,:,:]
    fact[fact<=-1] = -1
    fact[fact>=1] = 1
    fact = np.abs(fact)
    return np.multiply(field,fact, out = out) 


def project_normalized_fft(field, dmat, window = None, ref = None, out = None):
    if ref is not None:
        fref = fft2(ref, out = out)
        intensity1 = field2intensity(fref)
        f1 = fft2(field, out = out)
    else:
        f1 = fft2(field, out = out)
        intensity1 = field2intensity(f1)
    f2 = dotmf(dmat, f1 ,out = f1)
    intensity2 = field2intensity(f2)
    f3 = normalize_field(f2, intensity1, intensity2, out = f2)
    out = ifft2(f3, out = out)
    if window is not None:
        out = np.multiply(out,window,out = out)
    return out  

def transmitted_field_direct(field, beta, phi, n = 1.):
    ev = refind2eps([n]*3)
    ea = np.zeros_like(ev)
    alpha, fmat = alphaf(beta,phi, ev, ea)
    pmat = projection_mat(fmat)
    field0 = np.empty_like(field)
    dotmv(pmat,field2fvec(field), out = field2fvec(field0))
    return field0

def project_normalized_local(field, dmat, window = None, ref = None, out = None):
    f1 = fft2(field) 
    f2 = dotmf(dmat, f1 ,out = f1)
    f = ifft2(f2, out = f2)
    jmat = polarizer(jonesvec(field2fvec(f[...,::2,:,:])))
    #pmat1 = ray_jonesmat4x4(jmat)
    pmat1 = normal_polarizer(jonesvec(field2fvec(f[...,::2,:,:])))
    pmat2 = -pmat1
    pmat2[...,0,0] += 1
    pmat2[...,1,1] += 1 #pmat1 + pmat2 = identity by definition
    pmat2[...,2,2] += 1 #pmat1 + pmat2 = identity by definition
    pmat2[...,3,3] += 1 #pmat1 + pmat2 = identity by definition
    
    if ref is not None:
        intensity1 = field2intensity(dotmf(pmat1, ref))
        ref = dotmf(pmat2, ref)
        
    else:
        intensity1 = field2intensity(dotmf(pmat1, field))
        ref = dotmf(pmat2, field)

    intensity2 = field2intensity(f)
    
    f = normalize_field(f, intensity1, intensity2)
    
    out = np.add(f,ref,out = out)
    if window is not None:
        out = np.multiply(out,window,out = out)
    return out 

def transpose(field):
    """transposes field from shape (..., k,n,m) to (...,n,m,k). Inverse of
    itranspose_field"""
    taxis = list(range(field.ndim))
    taxis.append(taxis.pop(-3))
    return field.transpose(taxis) 

def normal_polarizer(jones = (1,0)):
    """A 4x4 polarizer for normal incidence light. It works reasonably well also
    for off-axis light, but it introduces weak reflections and depolarization.
    
    For off-axis planewaves you should use ray_polarizer instead of this."""
    p = polarizer(jonesvec(jones))
    pmat = np.zeros(shape = p.shape[:-2] + (4,4), dtype = p.dtype)
    pmat[...,::2,::2] = p
    pmat[...,1,1] = p[...,0,0]
    pmat[...,3,3] = p[...,1,1]
    pmat[...,1,3] = -p[...,0,1]
    pmat[...,3,1] = -p[...,1,0]
    return pmat

def project_normalized_local(field, dmat, window = None, ref = None, out = None):
    f1 = fft2(field) 
    f2 = dotmf(dmat, f1 ,out = f1)
    f = ifft2(f2, out = f2)
    pmat1 = normal_polarizer(jonesvec(transpose(f[...,::2,:,:])))
    pmat2 = -pmat1
    pmat2[...,0,0] += 1
    pmat2[...,1,1] += 1 #pmat1 + pmat2 = identity by definition
    pmat2[...,2,2] += 1 #pmat1 + pmat2 = identity by definition
    pmat2[...,3,3] += 1 #pmat1 + pmat2 = identity by definition
    
    if ref is not None:
        intensity1 = field2intensity(dotmf(pmat1, ref))
        ref = dotmf(pmat2, ref)
        
    else:
        intensity1 = field2intensity(dotmf(pmat1, field))
        ref = dotmf(pmat2, field)

    intensity2 = field2intensity(f)
    
    f = normalize_field(f, intensity1, intensity2)
    
    out = np.add(f,ref,out = out)
    if window is not None:
        out = np.multiply(out,window,out = out)
    return out 

def normalize_field_total(field, i1, i2, out = None):
    m = i2 == 0.
    i2[m] = 1.
    i1[m] = 0.
    fact = (i1/i2)
    fact[fact<=-1] = -1
    fact[fact>=1] = 1
    fact = np.abs(fact)
    fact = fact[...,None,None,None]
    return np.multiply(field,fact, out = out) 

def total_intensity(field):
    """Calculates total intensity of the field. 
    Computes intesity and sums over pixels."""
    i = field2intensity(field)
    return i.sum(tuple(range(i.ndim))[-2:])#sum over pixels

def project_normalized_total(field, dmat, window = None, ref = None, out = None):
    if ref is not None:
        i1 = total_intensity(ref)
    else:
        i1 = total_intensity(field)
    f1 = fft2(field, out = out)
    
    f2 = dotmf(dmat, f1 ,out = f1)
    out = ifft2(f2, out = out)
    i2 = total_intensity(out)
    
    out = normalize_field_total(out, i1, i2, out = out)
    
    if window is not None:
        out = np.multiply(out,window,out = out)
    return out    

def normalize_total(field, dmat, window = None, ref = None, out = None):
    if ref is not None:
        i1 = total_intensity(ref)
    else:
        i1 = total_intensity(field)
        
    f2 = dotmf(dmat, field, out = out)
    i2 = total_intensity(out)
    
    out = normalize_field_total(out, i1, i2, out = f2)
    
    if window is not None:
        out = np.multiply(out,window,out = out)
    return out  

def _projected_field(field, wavenumbers, mode, n = 1, betamax = BETAMAX,  out = None):
    eps = refind2eps([n]*3)
    pmat = projection_matrix(field.shape[-2:], wavenumbers, epsv = eps, epsa = (0.,0.,0.), mode = mode, betamax = betamax)
    return diffract(field, pmat, out = out) 

    
#def jones2H(jones,beta = 0., phi = 0., n = 1., out = None):
#    eps = refind2eps([n]*3)
#    layer = np.asarray((0.,0.,0.), dtype = FDTYPE)
#    alpha, f = alphaf(beta,phi,layer,eps) 
#    A = f[...,::2,::2]
#    B = f[...,1::2,::2]
#    Ai = inv(A)
#    D = dotmm(B,Ai)
#    return dotm1f(D,jones,out)

def jones2H(jones, wavenumbers, n = 1., betamax = BETAMAX, mode = +1, out = None):  
    eps = refind2eps([n]*3)
    shape = jones.shape[-2:]
    layer = np.asarray((0.,0.,0.), dtype = FDTYPE)
    alpha, f, fi = diffraction_alphaffi(shape, wavenumbers, epsv = eps, 
                            epsa = layer, betamax = betamax)
#    A = f[...,::2,::2]
#    B = f[...,1::2,::2]
#    Ai = inv(A)
#    D = dotmm(B,Ai)  
    D = E2H_mat(f, mode = mode)
    return diffract(jones, D, out = out) 


def transmitted_field(field, wavenumbers, n = 1, betamax = BETAMAX, out = None):
    """Computes transmitted (forward propagating) part of the field.
    
    Parameters
    ----------
    field : ndarray
        Input field array
    wavenumbers : array_like
        Wavenumbers of the field
    n : float, optional
        Refractive index of the media (1 by default)
    betamax : float, optional
        Betamax perameter used.
    out : ndarray, optinal
        Output array
        
    Returns
    -------
    out : ndarray
       Transmitted field.
    """
        
    return _projected_field(np.asarray(field), wavenumbers, +1, n = n, 
                            betamax = betamax, out = out) 
    
def reflected_field(field, wavenumbers, n = 1, betamax = BETAMAX,  out = None):
    """Computes reflected (backward propagating) part of the field.
    
    Parameters
    ----------
    field : ndarray
        Input field array
    wavenumbers : array_like
        Wavenumbers of the field
    n : float, optional
        Refractive index of the media (1 by default)
    betamax : float
        Betamax perameter used.
    norm : bool, optional
        Whether to normalize field so that power spectrum of the output field
        remains the same as that of the input field.
    out : ndarray, optinal
        Output array
        
    Returns
    -------
    out : ndarray
       Reflected field.
    """
    return _projected_field(np.asarray(field), wavenumbers, -1, n = n, betamax = betamax, out = out) 
    

def transfer_field(field_data, optical_data, beta = None, phi = None, nin = None, nout = None,  
           npass = None, nstep=1, diffraction = None, reflection = None, method = None, 
           multiray = False,
           norm = DTMM_NORM_FFT, betamax = BETAMAX, smooth = SMOOTH, split_rays = False,
           split_diffraction = False,split_wavelengths = False,
           eff_data = None, ret_bulk = False, out = None):
    """Tranfers input field data through optical data.
    
    This function calculates transmitted field and possibly (when npass > 1) 
    updates input field with reflected waves. 
    
    
    Parameters
    ----------
    field_data : Field data tuple
        Input field data tuple
    optical_data : Optical data tuple
        Optical data tuple through which input field is transfered.
    beta : float or 1D array_like of floats, optional
        Beta parameter of the input field. If it is a 1D array, beta[i] is the
        beta parameter of the field_data[0][i] field array.f not provided, beta
        is caluclated from input data (see also multiray option).
    phi : float or 1D array_like of floats, optional
        Phi angle of the input light field. If it is a 1D array, phi[i] is the
        phi parameter of the field_data[0][i] field array. If not provided, phi
        is caluclated from input data (see also multiray option).
    nin : float, optional
        Refractive index of the input (bottom) surface (1. by default). Used
        in combination with npass > 1 to determine reflections from input layer,
        or in combination with reflection = True to include Fresnel reflection
        from the input surface.
    nout : float, optional
        Refractive index of the output (top) surface (1. by default). Used
        in combination with npass > 1 to determine reflections from output layer,
        or in combination with reflection = True to include Fresnel reflection
        from the output surface.
    npass: int, optional
        How many passes (iterations) to perform. For strongly reflecting elements
        this should be set to a higher value. If npass > 1, then input field data is
        overwritten and adds reflected light from the sample (defaults to 1).
    nstep: int or 1D array_like of ints
        Specifies layer propagation computation steps (defaults to 1). For thick 
        layers you may want to increase this number. If layer thickness is greater
        than pixel size, you should increase this number.
    diffraction : bool or int, optional
        Defines how diffraction is calculated. Setting this to False or 0 will 
        disable diffraction calculation. Diffraction is enabled by default.
        If specified as an integer, it defines diffraction calculation quality.
        1 for simple (fast) calculation, higher numbers increase accuracy 
        and decrease computation speed. You can set it to np.inf or -1 for max 
        (full) diffraction calculation and very slow computation.
    reflection : bool or int or None, optional
        Reflection calculation mode for '2x2' method. It can be either
        0 or False for no reflections, 1 (default) for reflections in fft space 
        (from effective layers), or 2 for reflections in real space 
        (from individual layers). If this argument is not provided it is 
        automatically set to 0 if npass == 1 and 1 if npass > 1 and diffraction
        == 1 and to 2 if npass > 1 and diffraction > 1. See documentation for details. 
    method : str, optional
        Specifies which method to use, either '2x2' (default) or '4x4'.
    multiray : bool
        If specified it defines if first axis of the input data is treated as multiray data
        or not. If beta and phi are not set, you must define this if your data
        is multiray so that beta and phi values are correctly determined.
    norm : int, optional
        Normalization mode used when calculating multiple reflections with 
        npass > 1 and 4x4 method. Possible values are 0, 1, 2, default value is 1.
    smooth : float, optional
        Smoothing parameter when calculating multiple reflections with 
        npass > 1 and 4x4 method. Possible values are values above 0.Setting this
        to higher values > 1 removes noise but reduces convergence speed. Setting
        this to < 0.1 increases convergence, but it increases noise. 
    split_diffraction : bool, optional
        In diffraction > 1 calculation this option specifies whether to split 
        computation over single beam to consume less temporary memory storage.
        For large diffraction values this option should be set.
    split_rays: bool, optional
        In multi-ray computation this option specifies whether to split 
        computation over single rays to consume less temporary memory storage.
        For large multi-ray datasets this option should be set.
    split_wavelengths : bool, optional
        Specifies whether to treat data at each wavelength as an independent ray. 
        With off-axis propagation of eigenmodes with different beta values this
        should be set to true to assure proper determination of beta parameter
        for proper reflection calculation. Normally there will be very little
        difference, however, for small simulation volumes (compared to the 
        wavelength) and at large incidence angles, you should set this to True
        for simulations of multi-wavelength fields.
    eff_data : Optical data tuple or symmetry, optional
        Optical data tuple of homogeneous layers through which light is diffracted
        in the diffraction calculation when diffraction >= 1. If not provided, 
        an effective data is build from optical_data by taking the mean value 
        of the epsilon tensor. You can also provide the symmetry argument, e.g.
        'isotropic', 'uniaxial' or 'biaxial', or a list of these values specifying
        the symmetry of each of the layers. This argument is passed directly to 
        the :func:`.data.effective_data` function.
    ret_bulk : bool, optional
        Whether to return bulk field instead of the transfered field (default).
    out : ndarray, optional
        Output array.
    
    """
    nin = get_default_config_option("nin",nin)
    nout = get_default_config_option("nout",nout)
    method = get_default_config_option("method",method)
    npass = get_default_config_option("npass",npass)
    eff_data = get_default_config_option("eff_data",eff_data)
    diffraction = get_default_config_option("diffraction",diffraction)
    reflection = get_default_config_option("reflection",reflection)
    
    
    t0 = time.time()
    verbose_level = DTMMConfig.verbose
    
    
    if method == "4x4" and npass > 1 and diffraction == False:
        import warnings
        warnings.warn("The 4x4 method with diffraction disabled is not yet supported\
                      for input fields with beta >0. Use 2x2 method insted.")
        
    if npass == -1 or npass == np.inf:
        method = "4x4"
        diffraction = np.inf
        reflection = 2
        
    #choose best/supported reflection mode based on other arguments
    if reflection is None:
        reflection = 0 if method == "2x2" else 1
        if method == "4x4" and diffraction == 0:
            reflection = 2
        if npass > 1:
            reflection = 1
            if diffraction > 1 and method == "2x2":
                reflection = 2
                
    if verbose_level > 0:
        print("Transferring input field.")    
        
    eff_data_name = eff_data if eff_data in (0,1,2,"isotropic","uniaxial","biaxial") else "custom"
    
    if verbose_level > 1:
        print("------------------------------------")
        print(" $ calculation method: {}".format(method))  
        print(" $ reflection mode: {}".format(reflection)) 
        print(" $ diffraction mode: {}".format(diffraction))  
        print(" $ number of substeps: {}".format(nstep))     
        print(" $ input refractive index: {}".format(nin))   
        print(" $ output refractive index: {}".format(nout)) 
        print(" $ effective data: {}".format(eff_data_name)) 
        print(" $ max beta: {}".format(betamax)) 
        print("------------------------------------")
        
    
    
    field_in,wavelengths,pixelsize = field_data

#    if out is None:
#        if ret_bulk == True:
#            #must have a length of optical data + 2 extra layers
#            field_out = np.empty(shape = (len(optical_data[0])+2,)+field_in.shape, dtype = field_in.dtype)     
#        else:
#            field_out = np.empty_like(field_in) 
#    else:
#        field_out = out
    
    splitted_wavelengths = split_wavelengths == True and not isinstance(field_in, tuple) and ret_bulk == False

    
    if splitted_wavelengths:
        
        if out is None:
            out_field = np.empty_like(field_in) 
        else:
            out_field = out
        out = [out_field[...,i,:,:,:] for i in range(len(wavelengths))]
        field_in = tuple((field_in[...,i,:,:,:] for i in range(len(wavelengths))))

    if isinstance(field_in, tuple):
        nwavelengths = len(field_in)
        if out is None:
            out = [None for i in range(len(field_in))]
        for i,(f,w,o) in enumerate(zip(field_in, wavelengths, out)):
            if verbose_level >0:
                print("Wavelength {}/{}".format(i+1,nwavelengths))
            field_data = f,w,pixelsize
            o = _transfer_field(field_data, optical_data, beta, phi, nin, nout,  
                npass , nstep, diffraction, reflection , method, 
                multiray, norm, betamax, smooth, split_rays,
                split_diffraction, eff_data, ret_bulk, o) 
            out[i] = o
        out = tuple(out)
    else:
    
        out = _transfer_field(field_data, optical_data, beta, phi, nin, nout,  
               npass , nstep, diffraction, reflection , method, 
               multiray, norm, betamax, smooth, split_rays,
               split_diffraction ,
               eff_data, ret_bulk, out)   

    t = time.time()-t0
    if verbose_level >1:
        print("------------------------------------")
        print("   Done in {:.2f} seconds!".format(t))  
        print("------------------------------------")
    
    if splitted_wavelengths:
        return out_field, wavelengths, pixelsize
    else:
        return out

def _transfer_field(field_data, optical_data, beta, phi, nin, nout,  
           npass , nstep, diffraction, reflection , method, 
           multiray, norm, betamax, smooth, split_rays,
           split_diffraction ,
           eff_data, ret_bulk, out):
    verbose_level = DTMMConfig.verbose
 
    if split_rays == False:
        if method  == "4x4":
            if npass == -1 or npass == np.inf:
                out = transfer3d(field_data, optical_data,nin = nin, nout =nout, betamax = betamax)
            else:
                out = transfer_4x4(field_data, optical_data, beta = beta, 
                           phi = phi, eff_data = eff_data, nin = nin, nout = nout, npass = npass,nstep=nstep,
                      diffraction = diffraction, reflection = reflection, multiray = multiray,norm = norm, smooth = smooth,
                      betamax = betamax, ret_bulk = ret_bulk, out = out)
        else:
            out = transfer_2x2(field_data, optical_data, beta = beta, 
                   phi = phi, eff_data = eff_data, nin = nin, nout = nout, npass = npass,nstep=nstep,
              diffraction = diffraction,  multiray = multiray,split_diffraction = split_diffraction,reflection = reflection, betamax = betamax, ret_bulk = ret_bulk, out = out)
        
    else:#split input data by rays and compute ray-by-ray
        
        field_in,wavelengths,pixelsize = field_data
        if out is None:
            if ret_bulk == True:
                #must have a length of optical data + 2 extra layers
                field_out = np.empty(shape = (len(optical_data[0])+2,)+field_in.shape, dtype = field_in.dtype)     
            else:
                field_out = np.empty_like(field_in) 
        else:
            field_out = out
        nrays = len(field_in)
        if beta is None:
            beta = [None] * nrays
        if phi is None:
            phi = [None] * nrays
        multiray = False
        for i, bp in enumerate(zip(beta,phi)):
            if verbose_level >0:
                print("Ray {}/{}".format(i+1,nrays))
        
            field_data = (field_in[i],wavelengths, pixelsize)
            beta, phi = bp
            
            if ret_bulk == True:
                out = field_out[:,i]
            else:
                out = field_out[i]
            if method  == "4x4":
                 transfer_4x4(field_data, optical_data, beta = beta, 
                       phi = phi, eff_data = eff_data, nin = nin, nout = nout, npass = npass,nstep=nstep,
                  diffraction = diffraction, reflection = reflection,multiray = multiray,norm = norm, smooth = smooth,
                  betamax = betamax, out = out, ret_bulk = ret_bulk)
            else:
                transfer_2x2(field_data, optical_data, beta = beta, 
                   phi = phi, eff_data = eff_data, nin = nin, nout = nout, npass = npass,nstep=nstep,
              diffraction = diffraction,multiray = multiray, split_diffraction = split_diffraction, reflection = reflection, betamax = betamax, out = out, ret_bulk = ret_bulk)
        
            
        out = field_out, wavelengths, pixelsize
    return out
      

def transfer_4x4(field_data, optical_data, beta = 0., 
                   phi = 0., eff_data = None, nin = 1., nout = 1., npass = 1,nstep=1,
              diffraction = True, reflection = 1, multiray = False,norm = DTMM_NORM_FFT, smooth = SMOOTH,
              betamax = BETAMAX, ret_bulk = False, out = None):
    """Transfers input field data through optical data. See transfer_field.
    """
    if reflection not in (1,2,3,4):
        raise ValueError("Invalid reflection. The 4x4 method is either reflection mode 1 or 2.")
    if smooth > 1.:
        pass
        #smooth =1.
    elif smooth < 0:
        smooth = 0.
    verbose_level = DTMMConfig.verbose
    if verbose_level >1:
        print(" * Initializing.")
        
    calc_reference = bool(norm & DTMM_NORM_REF)
    
    if (norm & DTMM_NORM_FFT):
        norm = "fft"
    else:
        if calc_reference:
            norm = "local"
        else:
            norm = "total"
    
    #define optical data
    d, epsv, epsa = validate_optical_data(optical_data)
       
    layers, eff_layers = _layers_list(optical_data, eff_data, nin, nout, nstep)
            
    #define input field data
    field_in, wavelengths, pixelsize = field_data
    
    #define constants 
    ks = k0(wavelengths, pixelsize)
    
    n = len(layers)

    if beta is None and phi is None:
        ray_tracing = False
        beta, phi = field2betaphi(field_in,ks, multiray)
    else:
        ray_tracing = False
    if diffraction != 1:
        ray_tracing = False
    beta, phi = _validate_betaphi(beta,phi,extendeddim = field_in.ndim-2)
    
    #define output field
    if out is None:
        if ret_bulk == True:
            bulk_out = np.zeros((n,)+field_in.shape, field_in.dtype)
            bulk_out[0] = field_in
            field_in = bulk_out[0]
            field_out = bulk_out[-1]
        else:
            bulk_out = None
            field_out = np.zeros_like(field_in)   
    else:
        
        #out[...] = 0.
        if ret_bulk == True:
            bulk_out = out
            bulk_out[0] = field_in
            field_in = bulk_out[0]
            field_out = bulk_out[-1]
        else:
            bulk_out = None
            field_out = out
        #make sure we remove forward propagating waves
        reflected_field(field_out, ks, n = nin, betamax = betamax, out = field_out)
    indices = list(range(1,n-1))
    #if npass > 1:
    #    field0 = field_in.copy()
 
    #field0 = field_in.copy()
    field0 = transmitted_field(field_in, ks, n = nin, betamax = betamax)
    field = field_in.copy()

    
    #field_in[...] = 0.
            
    if norm == 2:
        #make sure we take only the forward propagating part of the field
        transmitted_field(field, ks, n = nin, betamax = min(betamax,nin), out = field)
        field_in[...] = field
        
    if calc_reference:
        ref = field.copy()
    else:
        ref = None    
    
    #i0 = field2intensity(transmitted_field(field0, ks, n = nin, betamax = betamax))
    i0 = field2intensity(field0)
    i0 = i0.sum(tuple(range(i0.ndim))[-2:]) 
    
    if reflection not in (2,4) and 0<= diffraction and diffraction < np.inf:
        work_in_fft = True
    else:
        work_in_fft = False
        
    if work_in_fft:
        field = fft2(field,out = field)
    _reuse = False
    tmpdata = {}
    
    #:projection matrices.. set when needed
    if npass > 1:
        pin_mat = projection_matrix(field.shape[-2:], ks,  epsv = refind2eps([nin]*3), mode = +1, betamax = betamax)
        pout_mat = projection_matrix(field.shape[-2:], ks,  epsv = refind2eps([nout]*3), mode = +1, betamax = betamax)

    
    for i in range(npass):
        if verbose_level > 0:
            prefix = " * Pass {:2d}/{}".format(i+1,npass)
            suffix = ""
        else:
            prefix = ""
            suffix = "{}/{}".format(i+1,npass)

        _bulk_out = bulk_out
        direction = (-1)**i 
        _nstep, (thickness,ev,ea)  = layers[indices[0]]
        
        if direction == 1:
            _betamax = betamax
        else:
            _betamax = betamax
        
        for pindex, j in enumerate(indices):
            print_progress(pindex,n,level = verbose_level, suffix = suffix, prefix = prefix) 
            
            nstep, (thickness,ev,ea) = layers[j]
            output_layer = (thickness*direction,ev,ea)

            _nstep, output_layer_eff = eff_layers[j]

            d,e,a = output_layer_eff
            output_layer_eff = d*direction, e,a 
            
            if ray_tracing == True:
                if work_in_fft:
                    beta, phi = field2betaphi(ifft2(field),ks, multiray)
                else:
                    beta, phi = field2betaphi(field,ks, multiray)
                beta, phi = _validate_betaphi(beta,phi,extendeddim = field_in.ndim-2)
            
            if calc_reference and i%2 == 0:
                _nstep, (thickness_in,ev_in,ea_in) = layers[j-1]
                input_layer = (thickness_in,ev_in,ea_in)
                ref2, refl = propagate_2x2_effective_2(ref[...,::2,:,:], ks, input_layer, output_layer ,None, output_layer_eff, 
                            beta = beta, phi = phi, nsteps = nstep, diffraction = diffraction, reflection = 0, 
                            betamax = _betamax,mode = direction, out = ref[...,::2,:,:])
                
            if bulk_out is not None:
                out_field = _bulk_out[j]
            else:
                out_field = field
            
            if diffraction >= 0 and diffraction < np.inf:
                if reflection == 4:
                    field = propagate_4x4_effective_4(field, ks, output_layer,output_layer_eff, 
                                beta = beta, phi = phi, nsteps = nstep, diffraction = diffraction, 
                                betamax = _betamax, out = out_field)  
                elif reflection == 3:
                    field = propagate_4x4_effective_3(field, ks, output_layer,output_layer_eff, 
                                beta = beta, phi = phi, nsteps = nstep, diffraction = diffraction, 
                                betamax = _betamax, out = out_field) 
                
                elif reflection ==2:
                    field = propagate_4x4_effective_2(field, ks, output_layer,output_layer_eff, 
                                beta = beta, phi = phi, nsteps = nstep, diffraction = diffraction, 
                                betamax = _betamax, out = out_field, tmpdata = tmpdata)
                else:
                    field = propagate_4x4_effective_1(field, ks, output_layer,output_layer_eff, 
                                beta = beta, phi = phi, nsteps = nstep, diffraction = diffraction, 
                                betamax = _betamax, out = out_field, _reuse = _reuse)                    
            else:
                field = propagate_4x4_full(field, ks, output_layer, 
                            nsteps = nstep, 
                            betamax = _betamax, out = out_field)
            _reuse = True
        if ref is not None:
            ref[...,1::2,:,:] = jones2H(ref2,ks,betamax = _betamax, n = nout)
        print_progress(n,n,level = verbose_level, suffix = suffix, prefix = prefix) 
        
        indices.reverse()
        
        if work_in_fft == True:
            field = ifft2(field)

        
        if npass > 1:
            #smooth = 1. - i/(npass-1.)
            #even passes - normalizing output field
            if i%2 == 0:
                if i == 0:
                    np.subtract(field,field_out, field)
                if i != npass -1:
                    if verbose_level > 1:
                        print(" * Normalizing transmissions.")
                    if calc_reference:
                        np.multiply(ref, (nin/nout)**0.5,ref)

                    sigma = smooth*(npass-i)/(npass)
                    
                    if norm == "fft":
                        field = project_normalized_fft(field, pout_mat, ref = ref, out = field)
                    elif norm == "local":
                        field = project_normalized_local(field, pout_mat, ref = ref, out = field)
                    elif norm == "total":
                        field = project_normalized_total(field, pout_mat, ref = ref, out = field)                      
                    
                    field = denoise_field(field, ks, nin, sigma, out = field)
                    
                np.add(field_out, field, field_out)
                field = field_out.copy()
                
            #odd passes - normalizeing input field   
            else:
                field_in[...] = field
                #if not the last pass...
                if i != npass -1:
                    if verbose_level > 1:
                        print(" * Normalizing reflections.")
                     
                    sigma = smooth*(npass-i)/(npass) 
                    
                    ffield = fft2(field, out = field)
                    ffield = dotmf(pin_mat, ffield, out = ffield)                                        
                    ffield = denoise_fftfield(ffield, ks, nin, sigma, out = ffield)
                    field = ifft2(field, out = ffield)
                    
                    i0f = total_intensity(field)
                    fact = ((i0/i0f))#**0.5)
                    fact = fact[...,None,None,None]
                    np.multiply(field,fact, out = field) 
                        
                    np.multiply(field_out,fact,out = field_out) 
                    np.multiply(field_in,fact,out = field_in) 
                        
                    np.subtract(field0,field, out = field)
                    
                    np.add(field_in,field,field_in)
                                                 
                    if calc_reference:
                        ref = field.copy()

            if work_in_fft == True:
                field = fft2(field, out = field)
                        
        else:
            field_out[...] = field
            #field_in[...] = field0
    #denoise(field_out, ks, nout, smooth*10, out = field_out)           
        
    if ret_bulk == True:
        if work_in_fft:
            ifft2(bulk_out[1:-1],out =bulk_out[1:-1])
        return bulk_out, wavelengths, pixelsize
    else:
        return field_out, wavelengths, pixelsize


    
    
    #m = tukey(beta,0.3,betamax)
#    #mask0 = (beta>0.9) & (beta < 1.1)
#    
#
#    
#    #mask = np.empty(mask0.shape + (4,), mask0.dtype)
#    #for i in range(4):
#    #    mask[...,i] = mask0   
#
#     
#    alpha, f, fi = alphaffi(beta,phi,epsv,epsa,out = out) 
#
#    out = (alpha,f,fi)
#    
#    try:
#        b1,betamax = betamax
#        a = (betamax-b1)/(betamax)
#        m = tukey(beta,a,betamax)
#        #print(b1)
#        #m = (b1-beta.clip(b1,betamax))/(betamax-b1)+1
#        #np.multiply(alpha,m[...,None,None],alpha)
#        np.multiply(f,m[...,None,None],f)    
#    
#                                
#    t = field2intensity(field)
#    t2 = ndimage.gaussian_filter(t, sigma)
#    
#    f = (dtmp+np.abs(t2))/(dtmp+np.abs(t))
#    if out is not None:
#        out[...] = field*f[...,None,:,:]
#        return out
#    else:
#        return field*f[...,None,:,:]




def _layers_list(optical_data, eff_data, nin, nout, nstep):
    """Build optical data layers list and effective data layers list.
    It appends/prepends input and output layers. A layer consists of
    a tuple of (n, thickness, epsv, epsa) where n is number of sublayers"""
    d, epsv, epsa = validate_optical_data(optical_data)  
    
    if epsa is not None:
        substeps = np.broadcast_to(np.asarray(nstep),(len(d),))
        layers = [(n,(t/n, ev, ea)) for n,t,ev,ea in zip(substeps, d, epsv, epsa)]
        #add input and output layers
        layers.insert(0, (1,(0., np.broadcast_to(refind2eps([nin]*3), epsv[0].shape), np.broadcast_to(np.array((0.,0.,0.), dtype = FDTYPE), epsa[0].shape))))
        layers.append((1,(0., np.broadcast_to(refind2eps([nout]*3), epsv[0].shape), np.broadcast_to(np.array((0.,0.,0.), dtype = FDTYPE), epsa[0].shape))))

        try:
            d_eff, epsv_eff, epsa_eff = validate_optical_data(eff_data, homogeneous = True)        
        except (TypeError, ValueError):
            if eff_data is None:
                d_eff, epsv_eff, epsa_eff = _isotropic_effective_data(optical_data)
            else:
                d_eff, epsv_eff, epsa_eff = effective_data(optical_data, symmetry = eff_data)
                        
        eff_layers = [(n,(t/n, ev, ea)) for n,t,ev,ea in zip(substeps, d_eff, epsv_eff, epsa_eff)]
        eff_layers.insert(0, (1,(0., refind2eps([nin]*3), np.array((0.,0.,0.), dtype = FDTYPE))))
        eff_layers.append((1,(0., refind2eps([nout]*3), np.array((0.,0.,0.), dtype = FDTYPE))))
        return layers, eff_layers
    else:
        substeps = np.broadcast_to(np.asarray(nstep),(len(d),))
        layers = [(n,(t/n, ev, None)) for n,t,ev in zip(substeps, d, epsv)]
        #add input and output layers
        layers.insert(0, (1,(0., np.broadcast_to(refind2eps([nin,nin,nin,0,0,0]), epsv[0].shape), None)))
        layers.append((1,(0., np.broadcast_to(refind2eps([nout,nout,nout,0,0,0]), epsv[0].shape), None)))

        try:
            d_eff, epsv_eff, epsa_eff = validate_optical_data(eff_data, homogeneous = True)        
        except (TypeError,ValueError):
            if eff_data is None:
                d_eff, epsv_eff, epsa_eff = _isotropic_effective_data(optical_data)
            else:
                d_eff, epsv_eff, epsa_eff = effective_data(optical_data, symmetry = eff_data)
                    
        eff_layers = [(n,(t/n, ev, ea)) for n,t,ev,ea in zip(substeps, d_eff, epsv_eff, epsa_eff)]
        eff_layers.insert(0, (1,(0., refind2eps([nin]*3), np.array((0.,0.,0.), dtype = FDTYPE))))
        eff_layers.append((1,(0., refind2eps([nout]*3), np.array((0.,0.,0.), dtype = FDTYPE))))
        return layers, eff_layers       


def transfer_2x2(field_data, optical_data, beta = None, 
                   phi = None, eff_data = None, nin = 1., 
                   nout = 1., npass = 1,nstep=1,
              diffraction = True, reflection = True, multiray = False, split_diffraction = False,
              betamax = BETAMAX, ret_bulk = False, out = None):
    """Tranfers input field data through optical data using the 2x2 method
    See transfer_field for documentation.
    """
    if reflection not in (0,1,2):
        raise ValueError("Invalid reflection. The 2x2 method supports reflection mode 0,1 or 2.")
    
    
    verbose_level = DTMMConfig.verbose
    if verbose_level >1:
        print(" * Initializing.")
    
    #create layers lists
    layers, eff_layers = _layers_list(optical_data, eff_data, nin, nout, nstep)
    #define input field data
    field_in, wavelengths, pixelsize = field_data
    #wavenumbers
    ks = k0(wavelengths, pixelsize)
    n = len(layers) - 1#number of interfaces
    
    if beta is None and phi is None:
        ray_tracing = False
        beta, phi = field2betaphi(field_in,ks, multiray)
    else:
        ray_tracing = False
    if diffraction != 1:
        ray_tracing = False
    beta, phi = _validate_betaphi(beta,phi,extendeddim = field_in.ndim-2)

    
    #define output field
    if out is None:
        if ret_bulk == True:
            bulk_out = np.zeros((n+1,)+field_in.shape, field_in.dtype)
            bulk_out[0] = field_in
            field_in = bulk_out[0]
            field_out = bulk_out[-1]
        else:
            bulk_out = None
            field_out = np.zeros_like(field_in)   
    else:
        out[...] = 0.
        if ret_bulk == True:
            bulk_out = out
            bulk_out[0] = field_in
            field_in = bulk_out[0]
            field_out = bulk_out[-1]
        else:
            bulk_out = None
            field_out = out
             
    indices = list(range(n))
        
    #make sure we take only the forward propagating part of the field
    if diffraction > 0:
        field0 = transmitted_field(field_in, ks, n = nin, betamax = betamax)
    else:
        field0 = transmitted_field_direct(field_in, beta, phi, n = nin)
    
    field = field0[...,::2,:,:].copy()
    
    if npass > 1:
        if reflection == 0:
            raise ValueError("Reflection mode `0` not compatible with npass > 1")
        field_in[...] = field0 #modify input field so that it has no back reflection
        #keep reference to reflected waves
        refl = np.zeros(shape = (n+1,) + field.shape[:-3] + (2,) + field.shape[-2:], dtype = field.dtype)
    else:
        #no need to store reflected waves
        refl = [None]*n
        
    if reflection in (0,1) and 0< diffraction and diffraction < np.inf:
        work_in_fft = True
    elif reflection == 1:
        work_in_fft = True
    else:   
        work_in_fft = False
        
    if work_in_fft:
        field = fft2(field,out = field)
        
    tmpdata = {}

    for i in range(npass):
        if verbose_level > 0:
            prefix = " * Pass {:2d}/{}".format(i+1,npass)
            suffix = ""
        else:
            prefix = ""
            suffix = "{}/{}".format(i+1,npass)
            
        if i > 0:
            field[...] = 0.
       
        direction = (-1)**i 
        _nstep, (thickness,ev,ea)  = layers[indices[0]]

        for pindex, j in enumerate(indices):
            print_progress(pindex,n,level = verbose_level, suffix = suffix, prefix = prefix) 
            
            jin = j+(1-direction)//2
            jout = j+(1+direction)//2
            _nstep, (thickness,ev,ea) = layers[jin]
            input_layer = (thickness,ev,ea)
            nstep, (thickness,ev,ea) = layers[jout]
            output_layer = (thickness*direction,ev,ea)

            _nstep, input_layer_eff = eff_layers[jin]
            _nstep, output_layer_eff = eff_layers[jout]

            d,e,a = input_layer_eff
            input_layer_eff = d*direction, e,a
            d,e,a = output_layer_eff
            output_layer_eff = d*direction, e,a     
        
            
            if jout == 0:
                bulk = field_in
            elif jout == len(indices):
                bulk = field_out
            else:
                if bulk_out is None:
                    bulk = None
                else:
                    bulk = bulk_out[jout]
                    
            if ray_tracing == True:
                if work_in_fft:
                    beta, phi = field2betaphi(ifft2(field),ks, multiray)
                else:
                    beta, phi = field2betaphi(field,ks, multiray)
                beta, phi = _validate_betaphi(beta,phi,extendeddim = field_in.ndim-2)

            if diffraction >= 0 and diffraction < np.inf:
     

                if work_in_fft:
                    field, refli = propagate_2x2_effective_1(field, ks, input_layer, output_layer ,input_layer_eff, output_layer_eff, 
                            beta = beta, phi = phi, nsteps = nstep, diffraction = diffraction, split_diffraction = split_diffraction, reflection = reflection, 
                            betamax = betamax,mode = direction, refl = refl[j], bulk = bulk, tmpdata = tmpdata)
                
                else:
                    field, refli = propagate_2x2_effective_2(field, ks, input_layer, output_layer ,input_layer_eff, output_layer_eff, 
                            beta = beta, phi = phi, nsteps = nstep, diffraction = diffraction, split_diffraction = split_diffraction, reflection = reflection, 
                            betamax = betamax,mode = direction, refl = refl[j], bulk = bulk, tmpdata = tmpdata)
                

            else:
                field, refli = propagate_2x2_full(field, ks, output_layer, input_layer = input_layer, 
                    nsteps = 1,  reflection = reflection, mode = direction,
                    betamax = betamax, refl = refl[j], bulk = bulk)

        print_progress(n,n,level = verbose_level, suffix = suffix, prefix = prefix) 
        
        indices.reverse()

    if ret_bulk == True:
        return bulk_out, wavelengths, pixelsize
    else:
        return field_out, wavelengths, pixelsize  
    
 
__all__ = ["transfer_field", "transmitted_field", "reflected_field", "transfer_2x2", "transfer_4x4", "total_intensity"]



"""
The 4x4 layer propagation functions
"""
from __future__ import absolute_import, print_function, division

from dtmm.conf import BETAMAX
from dtmm.wave import eigenwave, betaphi
from dtmm.tmm import alphaffi, phasem,  alphaf,  E_mat
from dtmm.linalg import dotmf, dotmdmf, inv, dotmdm,dotmf
from dtmm.diffract import diffract
from dtmm.fft import fft2, ifft2
import numpy as np
from dtmm.mode import fft_mask
from dtmm.matrix import corrected_Epn_diffraction_matrix, corrected_field_diffraction_matrix, \
         first_corrected_Epn_diffraction_matrix, second_corrected_Epn_diffraction_matrix, \
         first_field_diffraction_matrix, second_field_diffraction_matrix

def _transfer_ray_4x4_2(field, wavenumbers, layer,  beta = 0, phi=0,
                    nsteps = 1, dmatpn = None,
                    out = None, tmpdata = None):
    _out = {} if tmpdata is None else tmpdata
        
    d, epsv, epsa = layer
    
    if dmatpn is None:
        kd = wavenumbers*d
    else:
        dmatp, dmatn = dmatpn
        kd = wavenumbers*d/2    

    alpha, f, fi = alphaffi(beta,phi,epsv,epsa, out = _out.get("affi"))

    p = phasem(alpha,kd[...,None,None], out = _out.get("p"))
    if tmpdata is not None:
        tmpdata["affi"] = (alpha, f, fi)
        tmpdata["p"] = p
    
    if dmatpn is not None:
        e = E_mat(f, mode = None)
        ei = inv(e, out = _out.get("ei"))
        if tmpdata is not None:
            tmpdata["ei"] = ei
          
    for j in range(nsteps):
        if dmatpn is None:
            field = dotmdmf(f,p,fi,field, out = out)  
        else:
            field = dotmdmf(e,p,fi,field, out = out) 
            diffract(field[...,0::2,:,:], dmatp, out = field[...,0::2,:,:])
            diffract(field[...,1::2,:,:], dmatn, out = field[...,1::2,:,:])
            field = dotmdmf(f,p,ei,field, out = field)              
         
    return field

def _transfer_ray_4x4_4(field, wavenumbers, layer,  beta = 0, phi=0,
                    nsteps = 1, dmat = None,
                    out = None):

    d, epsv, epsa = layer
    
    if dmat is None:
        kd = wavenumbers*d
    else:
        kd = wavenumbers*d/2    

    alpha, f, fi = alphaffi(beta,phi,epsv,epsa)
    p = phasem(alpha,kd[...,None,None])
  
    for j in range(nsteps):
        if dmat is None:
            field = dotmdmf(f,p,fi,field, out = out)  
        else:
            field = dotmdmf(f,p,fi,field, out = out) 
            field = diffract(field, dmat, out = field)
            field = dotmdmf(f,p,fi,field, out = field)              
         
    return field


def _transfer_ray_4x4_1(field, wavenumbers, layer, dmat1, dmat2, beta = 0, phi=0,
                    nsteps = 1, 
                    betamax = BETAMAX, out = None):
        
    d, epsv, epsa = layer

    kd = wavenumbers*d 

    alpha, f = alphaf(beta,phi,epsv,epsa)
    p = phasem(alpha,kd[...,None,None])
    
    e = E_mat(f, mode = None)
    ei = inv(e)

    for j in range(nsteps):
        field = dotmf(dmat1,field, out = out)
        field = ifft2(field, out = field)
        field = dotmdmf(e,p,ei,field, out = field)  
        field = fft2(field, out = field)
        field = dotmf(dmat2,field, out = out)
                  
    return field

def _transfer_ray_4x4_1_windows(windows, field, wavenumbers, layer, dmat1, dmat2, betas, phis,
                    nsteps = 1, out = None,
                    betamax = BETAMAX):
        
    d, epsv, epsa = layer

    kd = wavenumbers*d 
    
    if out is None:
        out = np.zeros_like(field)

    for j in range(nsteps):
        field = dotmf(dmat1,field)
        for window, beta, phi in zip(windows, betas, phis):
            alpha, f = alphaf(beta,phi,epsv,epsa)
            p = phasem(alpha,kd[...,None,None])
            
            e = E_mat(f, mode = None)
            ei = inv(e)
            
            f = field * window
            
            
            f = ifft2(f, out = f)
            f = dotmdmf(e,p,ei,f, out = f)  
            f = fft2(field, out = f)
            out += f
            
        out = dotmf(dmat2,out, out = out)
                  
    return out

def _transfer_ray_4x4_3(field, wavenumbers, layer, dmat1, dmat2, beta = 0, phi=0,
                    nsteps = 1, 
                    betamax = BETAMAX, out = None):
        
    d, epsv, epsa = layer

    kd = wavenumbers*d 

    alpha, f, fi = alphaffi(beta,phi,epsv,epsa)
    p = phasem(alpha,kd[...,None,None])

    for j in range(nsteps):
        field = dotmf(dmat1,field, out = out)
        field = ifft2(field, out = field)
        field = dotmdmf(f,p,fi,field, out = field)  
        field = fft2(field, out = field)
        field = dotmf(dmat2,field, out = out)
                  
    return field


def propagate_4x4_effective_2(field, wavenumbers, layer, effective_layer, beta = 0, phi=0,
                    nsteps = 1, diffraction = True, 
                    betamax = BETAMAX,out = None, tmpdata = None):
    
    d_eff, epsv_eff, epsa_eff = effective_layer
    shape = field.shape[-2:]
    if diffraction <= 1:
        
        if diffraction != 0:

            dmatpn = corrected_Epn_diffraction_matrix(shape, wavenumbers, beta,phi, d = d_eff,
                                 epsv = epsv_eff, epsa = epsa_eff, betamax = betamax)
        
        else:
            dmatpn = None

        return _transfer_ray_4x4_2(field, wavenumbers, layer,
                                beta = beta, phi = phi, nsteps =  nsteps, dmatpn = dmatpn,
                                out = out, tmpdata = tmpdata)
    else:
        fout = np.zeros_like(field)
        _out = None
        field = fft2(field)

        try: 
            broadcast_shape = beta.shape
            beta = beta[...,0]
            phi = phi[...,0]
        except IndexError:
            broadcast_shape = ()
            
        windows, (betas, phis) = fft_mask(field.shape, wavenumbers, int(diffraction), 
                 betax_off = beta*np.cos(phi), betay_off = beta*np.sin(phi), betamax = betamax)    

        n = len(windows)
        betas = betas.reshape((n,) + broadcast_shape)
        phis = phis.reshape((n,) + broadcast_shape)
        
        dmatps, dmatns = corrected_Epn_diffraction_matrix(field.shape[-2:], wavenumbers, betas,phis, d = d_eff,
                                 epsv = epsv_eff, epsa = epsa_eff, betamax = betamax)


      
        for window, b, p, dmatp, dmatn  in zip(windows, betas, phis, dmatps,dmatns):
            fpart = np.multiply(field, window, out = _out)
            fpart_re = ifft2(fpart, out = fpart)

            _out =  _transfer_ray_4x4_2(fpart_re, wavenumbers, layer, 
                                beta = b, phi = p, nsteps =  nsteps,
                                dmatpn = (dmatp,dmatn), tmpdata = tmpdata)                       
            fout = np.add(fout, _out, out = fout)

        
#        for window, b, p  in zip(windows, betas, phis):
#            fpart = field * window
#            fpart_re = ifft2(fpart)
#            
#            b = b.reshape(beta_shape)
#            p = p.reshape(beta_shape)
#            
#            if diffraction != 0:
#                dmat = corrected_field_diffraction_matrix(field.shape[-2:], wavenumbers, b,p, d=d_eff,
#                                     epsv = epsv_eff, epsa = epsa_eff)
#            else:
#                dmat = None
#
#            _out =  _transfer_ray_4x4_2(fpart_re, wavenumbers, layer, 
#                                beta = b, phi = p, nsteps =  nsteps,
#                                dmat = dmat,_reuse = _reuse)                       
#            fout += _out

        if out is not None:
            out[...] = fout
        else:
            out = fout
        return out

def propagate_4x4_effective_4(field, wavenumbers, layer, effective_layer, beta = 0, phi=0,
                    nsteps = 1, diffraction = True, 
                    betamax = BETAMAX,out = None):
    
    d_eff, epsv_eff, epsa_eff = effective_layer
    
    if diffraction <= 1:
        dmat = corrected_field_diffraction_matrix(field.shape[-2:], wavenumbers, beta,phi, d=d_eff,
                                 epsv = epsv_eff, epsa = epsa_eff, betamax = betamax) if diffraction != 0 else None
        return _transfer_ray_4x4_4(field, wavenumbers, layer,
                                beta = beta, phi = phi, nsteps =  nsteps, dmat = dmat,
                                out = out)
    else:
        fout = np.zeros_like(field)
        _out = None
        field = fft2(field)

        try: 
            broadcast_shape = beta.shape
            beta = beta[...,0]
            phi = phi[...,0]
        except IndexError:
            broadcast_shape = ()
            
        windows, (betas, phis) = fft_mask(field.shape, wavenumbers, int(diffraction), 
                 betax_off = beta*np.cos(phi), betay_off = beta*np.sin(phi), betamax = betamax)    

        n = len(windows)
        betas = betas.reshape((n,) + broadcast_shape)
        phis = phis.reshape((n,) + broadcast_shape)
        
        if diffraction != 0:
            dmats = corrected_field_diffraction_matrix(field.shape[-2:], wavenumbers, betas,phis, d=d_eff,
                                 epsv = epsv_eff, epsa = epsa_eff)
        else:
            dmats = [None]*n
      
        for window, b, p, dmat  in zip(windows, betas, phis, dmats):
            fpart = np.multiply(field, window, out = _out)
            fpart_re = ifft2(fpart, out = fpart)

            _out =  _transfer_ray_4x4_4(fpart_re, wavenumbers, layer, 
                                beta = b, phi = p, nsteps =  nsteps,
                                dmat = dmat)                       
            fout = np.add(fout, _out, out = fout)



        
#        for window, b, p  in zip(windows, betas, phis):
#            fpart = field * window
#            fpart_re = ifft2(fpart)
#            
#            b = b.reshape(beta_shape)
#            p = p.reshape(beta_shape)
#            
#            if diffraction != 0:
#                dmat = corrected_field_diffraction_matrix(field.shape[-2:], wavenumbers, b,p, d=d_eff,
#                                     epsv = epsv_eff, epsa = epsa_eff)
#            else:
#                dmat = None
#
#            _out =  _transfer_ray_4x4_2(fpart_re, wavenumbers, layer, 
#                                beta = b, phi = p, nsteps =  nsteps,
#                                dmat = dmat,_reuse = _reuse)                       
#            fout += _out

        if out is not None:
            out[...] = fout
        else:
            out = fout
        return out

def propagate_4x4_effective_1(field, wavenumbers, layer, effective_layer, beta = 0, phi=0,
                    nsteps = 1, diffraction = True, 
                    betamax = BETAMAX,out = None,_reuse = False ):
    d_eff, epsv_eff, epsa_eff = effective_layer

    
    if diffraction == 1:
        dmat1 = first_corrected_Epn_diffraction_matrix(field.shape[-2:], wavenumbers, beta, phi,d_eff/2, epsv = epsv_eff, 
                                        epsa =  epsa_eff,betamax = betamax) 
        dmat2 = second_corrected_Epn_diffraction_matrix(field.shape[-2:], wavenumbers, beta, phi,d_eff/2, epsv = epsv_eff, 
                                        epsa =  epsa_eff,betamax = betamax) 
        return _transfer_ray_4x4_1(field, wavenumbers, layer,dmat1, dmat2, 
                                beta = beta, phi = phi, nsteps =  nsteps, 
                                betamax = betamax,  out = out)
    elif diffraction > 1:
        fout = np.zeros_like(field)
        _out = None

        try: 
            broadcast_shape = beta.shape
            beta = beta[...,0]
            phi = phi[...,0]
        except IndexError:
            broadcast_shape = ()
            
        windows, (betas, phis) = fft_mask(field.shape, wavenumbers, int(diffraction), 
                 betax_off = beta*np.cos(phi), betay_off = beta*np.sin(phi), betamax = betamax)    

        n = len(windows)
        betas = betas.reshape((n,) + broadcast_shape)
        phis = phis.reshape((n,) + broadcast_shape)

        dmats1 = first_corrected_Epn_diffraction_matrix(field.shape[-2:], wavenumbers, betas, phis,d_eff/2, epsv = epsv_eff, 
                                        epsa =  epsa_eff,betamax = betamax) 
        dmats2 = second_corrected_Epn_diffraction_matrix(field.shape[-2:], wavenumbers, betas, phis,d_eff/2, epsv = epsv_eff, 
                                        epsa =  epsa_eff,betamax = betamax) 

        for window, beta, phi, dmat1, dmat2  in zip(windows, betas, phis, dmats1, dmats2):
            fpart = np.multiply(field, window, out = _out)
            
            _out =  _transfer_ray_4x4_1(fpart, wavenumbers, layer, dmat1,dmat2,
                                beta = beta, phi = phi, nsteps =  nsteps,
                                betamax = betamax, out = _out)                       
            fout = np.add(fout, _out, out = fout)


#        for window, b, p  in zip(windows, betas, phis):
#            fpart = np.multiply(field, window, out = _out)
#            
#            beta = b.reshape(broadcast_shape)
#            phi = p.reshape(broadcast_shape)
#
#            dmat1 = second_field_diffraction_matrix(field.shape[-2:], wavenumbers, beta, phi,d_eff/2, epsv = epsv_eff, 
#                                            epsa =  epsa_eff,betamax = betamax) 
#            dmat2 = first_field_diffraction_matrix(field.shape[-2:], wavenumbers, beta, phi,d_eff/2, epsv = epsv_eff, 
#                                            epsa =  epsa_eff,betamax = betamax) 
#
#            _out =  _transfer_ray_4x4_1(fpart, wavenumbers, layer, dmat1,dmat2,
#                                beta = beta, phi = phi, nsteps =  nsteps,
#                                betamax = betamax, out = _out,_reuse = _reuse)                       
#            fout = np.add(fout, _out, out = fout)

        if out is not None:
            out[...] = fout
        else:
            out = fout
        return out
    else:
        raise ValueError("Invalid diffraction value")

def propagate_4x4_effective_1(field, wavenumbers, layer, effective_layer, beta = 0, phi=0,
                    nsteps = 1, diffraction = True, 
                    betamax = BETAMAX,out = None,_reuse = False ):
    d_eff, epsv_eff, epsa_eff = effective_layer

    
    if diffraction == 1:
        dmat1 = first_corrected_Epn_diffraction_matrix(field.shape[-2:], wavenumbers, beta, phi,d_eff/2, epsv = epsv_eff, 
                                        epsa =  epsa_eff,betamax = betamax) 
        dmat2 = second_corrected_Epn_diffraction_matrix(field.shape[-2:], wavenumbers, beta, phi,d_eff/2, epsv = epsv_eff, 
                                        epsa =  epsa_eff,betamax = betamax) 
        return _transfer_ray_4x4_1(field, wavenumbers, layer,dmat1, dmat2, 
                                beta = beta, phi = phi, nsteps =  nsteps, 
                                betamax = betamax,  out = out)
    elif diffraction > 1:
        fout = np.zeros_like(field)
        _out = None

        try: 
            broadcast_shape = beta.shape
            beta = beta[...,0]
            phi = phi[...,0]
        except IndexError:
            broadcast_shape = ()
            
        windows, (betas, phis) = fft_mask(field.shape, wavenumbers, int(diffraction), 
                 betax_off = beta*np.cos(phi), betay_off = beta*np.sin(phi), betamax = betamax)    

        n = len(windows)
        betas = betas.reshape((n,) + broadcast_shape)
        phis = phis.reshape((n,) + broadcast_shape)

        dmats1 = first_corrected_Epn_diffraction_matrix(field.shape[-2:], wavenumbers, betas, phis,d_eff/2, epsv = epsv_eff, 
                                        epsa =  epsa_eff,betamax = betamax) 
        dmats2 = second_corrected_Epn_diffraction_matrix(field.shape[-2:], wavenumbers, betas, phis,d_eff/2, epsv = epsv_eff, 
                                        epsa =  epsa_eff,betamax = betamax) 

        for window, beta, phi, dmat1, dmat2  in zip(windows, betas, phis, dmats1, dmats2):
            fpart = np.multiply(field, window, out = _out)
            
            _out =  _transfer_ray_4x4_1(fpart, wavenumbers, layer, dmat1,dmat2,
                                beta = beta, phi = phi, nsteps =  nsteps,
                                betamax = betamax, out = _out)                       
            fout = np.add(fout, _out, out = fout)


#        for window, b, p  in zip(windows, betas, phis):
#            fpart = np.multiply(field, window, out = _out)
#            
#            beta = b.reshape(broadcast_shape)
#            phi = p.reshape(broadcast_shape)
#
#            dmat1 = second_field_diffraction_matrix(field.shape[-2:], wavenumbers, beta, phi,d_eff/2, epsv = epsv_eff, 
#                                            epsa =  epsa_eff,betamax = betamax) 
#            dmat2 = first_field_diffraction_matrix(field.shape[-2:], wavenumbers, beta, phi,d_eff/2, epsv = epsv_eff, 
#                                            epsa =  epsa_eff,betamax = betamax) 
#
#            _out =  _transfer_ray_4x4_1(fpart, wavenumbers, layer, dmat1,dmat2,
#                                beta = beta, phi = phi, nsteps =  nsteps,
#                                betamax = betamax, out = _out,_reuse = _reuse)                       
#            fout = np.add(fout, _out, out = fout)

        if out is not None:
            out[...] = fout
        else:
            out = fout
        return out
    else:
        raise ValueError("Invalid diffraction value")


def propagate_4x4_effective_3(field, wavenumbers, layer, effective_layer, beta = 0, phi=0,
                    nsteps = 1, diffraction = True, 
                    betamax = BETAMAX,out = None):
    d_eff, epsv_eff, epsa_eff = effective_layer
    
    if diffraction == 1:
        dmat1 = second_field_diffraction_matrix(field.shape[-2:], wavenumbers, beta, phi,d_eff/2, epsv = epsv_eff, 
                                        epsa =  epsa_eff,betamax = betamax) 
        dmat2 = first_field_diffraction_matrix(field.shape[-2:], wavenumbers, beta, phi,d_eff/2, epsv = epsv_eff, 
                                        epsa =  epsa_eff,betamax = betamax) 
        return _transfer_ray_4x4_3(field, wavenumbers, layer,dmat1, dmat2, 
                                beta = beta, phi = phi, nsteps =  nsteps, 
                                betamax = betamax,  out = out)
    elif diffraction > 1:
        fout = np.zeros_like(field)
        _out = None

        try: 
            broadcast_shape = beta.shape
            beta = beta[...,0]
            phi = phi[...,0]
        except IndexError:
            broadcast_shape = ()
            
        windows, (betas, phis) = fft_mask(field.shape, wavenumbers, int(diffraction), 
                 betax_off = beta*np.cos(phi), betay_off = beta*np.sin(phi), betamax = betamax)    

        n = len(windows)
        betas = betas.reshape((n,) + broadcast_shape)
        phis = phis.reshape((n,) + broadcast_shape)

        dmats1 = second_field_diffraction_matrix(field.shape[-2:], wavenumbers, betas, phis,d_eff/2, epsv = epsv_eff, 
                                        epsa =  epsa_eff,betamax = betamax) 
        dmats2 = first_field_diffraction_matrix(field.shape[-2:], wavenumbers, betas, phis,d_eff/2, epsv = epsv_eff, 
                                        epsa =  epsa_eff,betamax = betamax) 

        for window, beta, phi, dmat1, dmat2  in zip(windows, betas, phis, dmats1, dmats2):
            fpart = np.multiply(field, window, out = _out)
            
            _out =  _transfer_ray_4x4_3(fpart, wavenumbers, layer, dmat1,dmat2,
                                beta = beta, phi = phi, nsteps =  nsteps,
                                betamax = betamax, out = _out)                       
            fout = np.add(fout, _out, out = fout)


#        for window, b, p  in zip(windows, betas, phis):
#            fpart = np.multiply(field, window, out = _out)
#            
#            beta = b.reshape(broadcast_shape)
#            phi = p.reshape(broadcast_shape)
#
#            dmat1 = second_field_diffraction_matrix(field.shape[-2:], wavenumbers, beta, phi,d_eff/2, epsv = epsv_eff, 
#                                            epsa =  epsa_eff,betamax = betamax) 
#            dmat2 = first_field_diffraction_matrix(field.shape[-2:], wavenumbers, beta, phi,d_eff/2, epsv = epsv_eff, 
#                                            epsa =  epsa_eff,betamax = betamax) 
#
#            _out =  _transfer_ray_4x4_1(fpart, wavenumbers, layer, dmat1,dmat2,
#                                beta = beta, phi = phi, nsteps =  nsteps,
#                                betamax = betamax, out = _out,_reuse = _reuse)                       
#            fout = np.add(fout, _out, out = fout)

        if out is not None:
            out[...] = fout
        else:
            out = fout
        return out

 
def propagate_4x4_full(field, wavenumbers, layer, 
                    nsteps = 1,  betamax = BETAMAX, out = None):

    shape = field.shape[-2:]

    d, epsv, epsa = layer

    kd = wavenumbers*d/nsteps
    
    if out is None:
        out = np.empty_like(field)
    
    out_af = None
    pm = None
    
    ii,jj = np.meshgrid(range(shape[0]), range(shape[1]),copy = False, indexing = "ij") 
    
    for step in range(nsteps):
        for i in range(len(wavenumbers)):
            ffield = fft2(field[...,i,:,:,:])
            ofield = np.zeros_like(out[...,i,:,:,:])
            b,p = betaphi(shape,wavenumbers[i])
            mask = b < betamax
            
            amplitude = ffield[...,mask]
            
            betas = b[mask]
            phis = p[mask]
            iind = ii[mask]
            jind = jj[mask]
              
            for bp in sorted(zip(range(len(betas)),betas,phis,iind,jind),key = lambda el : el[1], reverse = False):     
                        
                #for j,bp in enumerate(zip(betas,phis,iind,jind)):     
                              
                j, beta, phi, ieig, jeig = bp

                out_af = alphaffi(beta,phi,epsv,epsa, out = out_af)
                alpha,f,fi = out_af                      
                        
                pm = phasem(alpha,kd[i], out = pm)
                w = eigenwave(amplitude.shape[:-1]+shape, ieig,jeig, amplitude = amplitude[...,j])
                w = dotmdmf(f,pm,fi,w, out = w)
                #m = dotmdm(f,pm,fi)
                #w = dotmf(m,w, out = w)
                np.add(ofield,w,ofield)
                
            out[...,i,:,:,:] = ofield
        field = out
    return out



"""Rotation matrices and conversion functions.

Rotation matrices
-----------------

:func:`rotation_vector2` : 2D rotation vector.
:func:`rotation_matrix2` : 2D rotation matrix.
:func:`rotation_matrix_z` : 3D rotation matrix around z.
:func:`rotation_matrix_x` : 3D rotation matrix around x.
:func:`rotation_matrix_y` : 3D rotation matrix around y.
:func:`rotation_matrix` : general 3D rotation matrix from Euler angles.

Conversion functions
--------------------

:func:`rotate_diagonal_tensor` for 3D diagonal tensor rotation.
:func:`rotate_tensor` for 3D tensor rotation.
:func:`rotate_vector` for 3D vector rotation.
:func:`rotation_angles` fConverts rotation matrix to Euler angles

"""

from __future__ import absolute_import, print_function, division

import numpy as np
from math import cos, sin
from numba import jit

import numba as nb


from dtmm.conf import NCDTYPE, NFDTYPE, NF32DTYPE, NF64DTYPE, NC64DTYPE,NC128DTYPE, \
       CDTYPE, FDTYPE, NUMBA_TARGET, NUMBA_CACHE, NUMBA_FASTMATH


def rotation_vector2(angle, out=None):
    """
    Converts the provided angle into a rotation vector (cos, sin).

    Parameters
    ----------
    angle : array
         Array containing the angle of rotation at different points in space
    out : array
        Rotation, represented as a 2D vector at every point in space

    Returns
    -------

    """
    # calculate rotations from angle
    c, s = np.cos(angle), np.sin(angle)

    # Create <out> if not provided
    if out is None:
        out = np.empty(shape=c.shape + (2,), dtype=FDTYPE)

    # Store values
    out[..., 0] = c
    out[..., 1] = s

    return out

def rotation_matrix2(angle, out = None):
    """Returns 2D rotation matrix.
    
    Numpy broadcasting rules apply.
    """
    c,s = np.cos(angle), np.sin(angle)
    if out is None:
        out = np.empty(shape = c.shape + (2,2), dtype = FDTYPE)
    out[...,0,0] = c
    out[...,1,1] = c
    out[...,0,1] = -s
    out[...,1,0] = s    
    return out

def rotation_matrix_z(angle, out = None):
    """Calculates a rotation matrix for rotations around the z axis.
    
    Numpy broadcasting rules apply.
    """
    c,s = np.cos(angle), np.sin(angle)
    if out is None:
        out = np.zeros(shape = c.shape + (3,3), dtype = FDTYPE)
    out[...,0,0] = c
    out[...,0,1] = -s
    out[...,1,0] = s
    out[...,1,1] = c
    out[...,2,2] = 1.
    return out

def rotation_matrix_y(angle, out = None):
    """Calculates a rotation matrix for rotations around the y axis.
    
    Numpy broadcasting rules apply.
    """
    c,s = np.cos(angle), np.sin(angle)
    if out is None:
        out = np.zeros(shape = c.shape + (3,3), dtype = FDTYPE)
    out[...,0,0] = c
    out[...,0,2] = s
    out[...,1,1] = 1.
    out[...,2,0] = -s
    out[...,2,2] = c    
    return out

def rotation_matrix_x(angle, out = None):
    """Calculates a rotation matrix for rotations around the x axis.
    
    Numpy broadcasting rules apply.
    """
    c,s = np.cos(angle), np.sin(angle)
    if out is None:
        out = np.zeros(shape = c.shape + (3,3), dtype = FDTYPE)
    out[...,0,0] = 1.
    out[...,1,1] = c
    out[...,1,2] = -s
    out[...,2,1] = s
    out[...,2,2] = c    
    return out
           
@jit([(NF32DTYPE,NF32DTYPE,NF32DTYPE,NF32DTYPE[:,:]),
      (NF64DTYPE,NF64DTYPE,NF64DTYPE,NFDTYPE[:,:])],nopython = True, cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH) 
def _rotation_matrix(psi,theta,phi, R):
    """Fills rotation matrix values R = Rphi.Rtheta.Rpsi, where rphi and Rpsi are 
    rotations around y and Rtheta around z axis. 
    """
    cospsi = cos(psi)
    sinpsi = sin(psi)
    costheta = cos(theta)
    sintheta = sin(theta)
    cosphi = cos(phi)
    sinphi = sin(phi)

    sinphi_sinpsi = sinphi * sinpsi
    sinphi_cospsi = sinphi * cospsi    

    cosphi_sinpsi = cosphi * sinpsi
    cosphi_cospsi = cosphi * cospsi
    
    R[0,0] = costheta * cosphi_cospsi - sinphi_sinpsi
    R[0,1] = - costheta * cosphi_sinpsi - sinphi_cospsi
    R[0,2] = cosphi * sintheta
    R[1,0] = costheta * sinphi_cospsi + cosphi_sinpsi
    R[1,1] = cosphi_cospsi - costheta * sinphi_sinpsi
    R[1,2] = sintheta * sinphi
    R[2,0] = - cospsi * sintheta
    R[2,1] = sintheta*sinpsi
    R[2,2] = costheta
    
@jit([(NF32DTYPE[:,:],NF32DTYPE[:]),
      (NF64DTYPE[:,:],NFDTYPE[:])],nopython = True, cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)     
def _rotation_angles(R, out):
    """Computes the three Euler angles from the rotation matrix"""
    r22 = max(-1,min(1,R[2,2])) #round rotation matrix in case we have rounding issues in input matrix
    theta = np.arccos(r22)
    #if sin(theta) == 0., then R[1,2] and R[0,2] are zero
    if R[1,2] == 0. and R[0,2] == 0. or theta == 0.:
        #it does not matter what psi is, so set to zero
        psi = 0.
        r11 = max(-1,min(1,R[1,1])) #round rotation matrix in case we have rounding issues in input matrix
        
        #np.arccos(R[1,1]) is phi -psi, but since we set psi ti zero we may set this to phi.
        phi = np.arccos(r11)
    else:
        phi = np.arctan2(R[1,2],R[0,2])
        psi = np.arctan2(R[2,1],-R[2,0])
    out[0] = psi
    out[1] = theta
    out[2] = phi
    
    
@jit([(NF32DTYPE,NF32DTYPE,NF32DTYPE[:,:]),
      (NF64DTYPE,NF64DTYPE,NFDTYPE[:,:])],nopython = True, cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH) 
def _rotation_matrix_uniaxial(theta,phi, R):
    """Fills rotation matrix values R = Rphi.Rtheta, where phi is
    rotations around y and Rtheta around z axis. 
    """
    costheta = cos(theta)
    sintheta = sin(theta)
    cosphi = cos(phi)
    sinphi = sin(phi)
 
    R[0,0] = costheta * cosphi
    R[0,1] =  - sinphi 
    R[0,2] = cosphi * sintheta
    R[1,0] = costheta * sinphi  
    R[1,1] = cosphi
    R[1,2] = sintheta * sinphi
    R[2,0] = -sintheta
    R[2,1] = 0.
    R[2,2] = costheta
    

@nb.guvectorize([(NF32DTYPE[:],NF32DTYPE[:,:]),
                 (NF64DTYPE[:],NFDTYPE[:,:])], "(n)->(n,n)", 
                 target = NUMBA_TARGET, cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)
def rotation_matrix(angles, out):
    """rotation_matrix(angles, out)
    
    Calculates a general rotation matrix for rotations z-y-z psi, theta, phi.
    If out is specified.. it should be 3x3 float matrix.
    
    Parameters
    ----------
    angles : array_like
        A length 3 vector of the three angles
        
    Examples
    --------
    
    >>> a = rotation_matrix([0.12,0.245,0.78])
    
    The same can be obtained by:
        
    >>> Ry = rotation_matrix_z(0.12)
    >>> Rt = rotation_matrix_y(0.245)
    >>> Rf = rotation_matrix_z(0.78)
      
    >>> b = np.dot(Rf,np.dot(Rt,Ry))
    >>> np.allclose(a,b)
    True
    """   
    if len(angles) != 3:
        raise ValueError("Invalid input data shape")
    _rotation_matrix(angles[0],angles[1],angles[2], out)


@nb.guvectorize([(NF32DTYPE[:,:],NF32DTYPE[:]),
                 (NF64DTYPE[:,:],NFDTYPE[:])], "(n,n)->(n)", 
                 target = NUMBA_TARGET, cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)
def rotation_angles(matrix, out):
    """Computes Euler rotation angles from rotation matrix.
    
    Numpy broadcasting rules apply.
    
    Parameters
    ----------
    matrix : (..., 3,3) array
        Rotation matrix
    out : ndarray, optional
        Output array
    
    Returns
    -------
    angles : (...,3) ndarray
        Euler angles : psi (z rotation), theta (x rotation), phi (z rotation). 
    """
    
    if len(matrix) != 3:
        raise ValueError("Invalid input data shape")
    _rotation_angles(matrix, out)

def rotate_diagonal_tensor(R,diagonal,out = None):
    """Rotates a diagonal tensor, based on the rotation matrix provided
    
    >>> R = rotation_matrix((0.12,0.245,0.78))
    >>> diag = np.array([1.3,1.4,1.5], dtype = CDTYPE)
    >>> tensor = rotate_diagonal_tensor(R, diag)
    >>> matrix = dtmm.data.tensor2matrix(tensor)
    
    The same can be obtained by:

    >>> Ry = rotation_matrix_z(0.12)
    >>> Rt = rotation_matrix_y(0.245)
    >>> Rf = rotation_matrix_z(0.78)   
    >>> R = np.dot(Rf,np.dot(Rt,Ry)) 
    
    >>> diag = np.diag([1.3,1.4,1.5]) + 0j
    >>> matrix2 = np.dot(R,np.dot(diag, R.transpose()))
    
    >>> np.allclose(matrix2,matrix)
    True
    """
    diagonal = np.asarray(diagonal)
    tensor = np.empty(shape = diagonal.shape[:-1] + (6,), dtype = diagonal.dtype)
    tensor[...,0:3] = diagonal
    tensor[...,3:] = 0
    return rotate_tensor(R,tensor,out)


        
@jit([NCDTYPE[:](NFDTYPE[:,:],NCDTYPE[:],NCDTYPE[:])],nopython = True, cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)
def _rotate_diagonal_tensor(R,diagonal,out):
    """Calculates out = R.diagonal.RT of a diagonal tensor"""
    for i in range(3):
        out[i] = diagonal[0]*R[i,0]*R[i,0] + diagonal[1]*R[i,1]*R[i,1] + diagonal[2]*R[i,2]*R[i,2]
    out[3] = diagonal[0]*R[0,0]*R[1,0] + diagonal[1]*R[0,1]*R[1,1] + diagonal[2]*R[0,2]*R[1,2]
    out[4] = diagonal[0]*R[0,0]*R[2,0] + diagonal[1]*R[0,1]*R[2,1] + diagonal[2]*R[0,2]*R[2,2]          
    out[5] = diagonal[0]*R[1,0]*R[2,0] + diagonal[1]*R[1,1]*R[2,1] + diagonal[2]*R[1,2]*R[2,2]
    return out


@nb.guvectorize([(NFDTYPE[:,:],NCDTYPE[:],NCDTYPE[:])], "(m,m),(n)->(n)", 
                 target = NUMBA_TARGET, cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)
def rotate_tensor(R,tensor,out):
    """Calculates out = R.tensor.RT of a tensor"
    
    >>> R = rotation_matrix((0.12,0.245,0.78))
    >>> tensor = np.array([1.3,1.4,1.5,0.1,0.2,0.3], dtype = CDTYPE)
    >>> tensor = rotate_tensor(R, tensor)
    >>> matrix = tensor2matrix(tensor)
    """
    r11 = R[0,0]
    r12 = R[0,1]
    r13 = R[0,2]
    r21 = R[1,0]
    r22 = R[1,1]
    r23 = R[1,2]
    r31 = R[2,0]
    r32 = R[2,1]
    r33 = R[2,2]
    t1 = tensor[0]
    t2 = tensor[1]
    t3 = tensor[2]
    t4 = tensor[3]
    t5 = tensor[4]
    t6 = tensor[5]    
    
    a1 = (r11 * t1 + r12 * t4 + r13 * t5)
    a2 = (r13 * t3 + r11 * t5 + r12 * t6)
    a3 = (r12 * t2 + r11 * t4 + r13 * t6)
    
    out[0] = r11 * a1 + r13 * a2 + r12 * a3 
    out[3] = r21 * a1 + r23 * a2 + r22 * a3  
    out[4] = r31 * a1 + r33 * a2 + r32 * a3 
    
    a1 = r21 * t1 + r22 * t4 + r23 * t5 
    a2 = r23 * t3 + r21 * t5 + r22 * t6  
    a3 = r22 * t2 + r21 * t4 + r23 * t6
    
    out[1] =  r21 * a1 + r23 * a2 + r22 * a3
    out[5] =  r31 * a1 + r33 * a2 + r32 * a3
    
    out[2] = r31 * (r31 * t1 + r32 * t4 + r33 * t5) + r33 * (r33 * t3 + r31 * t5 + r32 * t6) + r32 * (r32 * t2 + r31 * t4 + r33 * t6)
    

@jit([(NF32DTYPE[:, :], NF32DTYPE[:], NF32DTYPE[:]),
      (NF64DTYPE[:, :], NF64DTYPE[:], NFDTYPE[:])], nopython=True, cache=NUMBA_CACHE, fastmath=NUMBA_FASTMATH)
def _rotate_vector(rotation_matrix, vector, out):

    # Rotate x values
    out0 = vector[0] * rotation_matrix[0, 0] + vector[1] * rotation_matrix[0, 1] + vector[2] * rotation_matrix[0, 2]
    # Rotate y values
    out1 = vector[0] * rotation_matrix[1, 0] + vector[1] * rotation_matrix[1, 1] + vector[2] * rotation_matrix[1, 2]
    # Rotate z values
    out2 = vector[0] * rotation_matrix[2, 0] + vector[1] * rotation_matrix[2, 1] + vector[2] * rotation_matrix[2, 2]

    # Set outputs, return by reference
    out[0] = out0
    out[1] = out1
    out[2] = out2


@nb.guvectorize([(NF32DTYPE[:, :], NF32DTYPE[:], NF32DTYPE[:]),
                 (NF64DTYPE[:, :], NF64DTYPE[:], NFDTYPE[:]), ], "(n,n),(n)->(n)", cache=NUMBA_CACHE,
                fastmath=NUMBA_FASTMATH, target=NUMBA_TARGET)
def rotate_vector(rotation_matrix, vector, out):
    """
    Rotates vector <vector> using rotation matrix <rotation_matrix>
    rotate_vector(R, vector)

    Calculates out = R.vector of a vector.

    Parameters
    ----------
    rotation_matrix : array
        3x3 rotation matrix
    vector : array
        Input 3-vector to rotate
    out : array, optional
        Output rotated 3-vector

    Returns
    -------
    vector : ndarray
        Rotated vector.
    """
    vector = np.asarray(vector)
    # Ensure size is correct size
    if len(vector) != 3:
        raise ValueError("Invalid input data shape")

    # Perform rotation
    _rotate_vector(rotation_matrix, vector, out)
    

@jit([NFDTYPE[:,:](NFDTYPE,NFDTYPE[:],NFDTYPE[:,:])],nopython = True, cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH) 
def _calc_rotations_uniaxial(phi0,element,R):
    theta = element[1]
    phi = element[2] -phi0
    _rotation_matrix_uniaxial(theta,phi, R)
    return R    


@jit([NFDTYPE[:,:](NFDTYPE,NFDTYPE[:],NFDTYPE[:,:])],nopython = True, cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH) 
def _calc_rotations(phi0,element,R):
    psi = element[0]
    theta = element[1]
    phi = element[2] -phi0
    _rotation_matrix(psi,theta,phi, R)
    return R   

if __name__ == "__main__":
    import doctest
    doctest.testmod()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Diffraction matrices used in the difraction step of the layer propagation functions
"""

from __future__ import absolute_import, print_function, division

from dtmm.conf import cached_function, BETAMAX,FDTYPE,CDTYPE
from dtmm.tmm import alphaffi, alphaEEi, alphaf,  E_mat, phase_mat
from dtmm.linalg import dotmdm, dotmm,  inv
from dtmm.diffract import diffraction_alphaffi, E_diffraction_matrix, phase_matrix, diffraction_alphaf
from dtmm.mode import fft_mask
import numpy as np
from dtmm.diffract import field_diffraction_matrix

@cached_function
def correction_matrix(beta,phi,ks, d=1., epsv = (1,1,1), epsa = (0,0,0.), out = None):
    alpha, f, fi = alphaffi(beta,phi,epsv,epsa)  
    kd = -np.asarray(ks)*d
    pmat = phase_matrix(alpha, kd)  
    return dotmdm(f,pmat,fi, out = out)

@cached_function
def field_correction_matrix(beta,phi,ks, d=1., epsv = (1,1,1), epsa = (0,0,0.), out = None):
    alpha, f, fi = alphaffi(beta,phi,epsv,epsa)  
    kd = -np.asarray(ks)*d
    pmat = phase_matrix(alpha, kd)  
    return dotmdm(f,pmat,fi, out = out)

@cached_function
def E_correction_matrix(beta,phi,ks, d=1., epsv = (1,1,1), epsa = (0,0,0.), mode = +1, out = None):
    alpha, j, ji = alphaEEi(beta,phi,epsv,epsa, mode = mode)  
    kd = -np.asarray(ks)*d
    pmat = phase_matrix(alpha, kd)  
    return dotmdm(j,pmat,ji, out = out)

@cached_function
def corrected_E_diffraction_matrix(shape, ks, beta,phi, d=1.,
                                 epsv = (1,1,1), epsa = (0,0,0.), mode = +1, betamax = BETAMAX, out = None):
    dmat = E_diffraction_matrix(shape, ks, d, epsv, epsa, mode = mode, betamax = betamax)
    cmat = E_correction_matrix(beta, phi, ks, d/2., epsv, epsa, mode = mode)
    return dotmm(cmat,dotmm(dmat,cmat, out = out), out = out)
    
 
@cached_function
def second_E_diffraction_matrix(shape, ks, beta,phi, d=1.,
                                 epsv = (1,1,1), epsa = (0,0,0.), mode = +1, betamax = BETAMAX, out = None):
    dmat = E_diffraction_matrix(shape, ks, d, epsv, epsa, mode = mode,betamax = betamax)
    cmat = E_correction_matrix(beta, phi, ks, d, epsv, epsa, mode = mode)
    return dotmm(dmat,cmat, out = None)

@cached_function
def first_E_diffraction_matrix(shape, ks, beta,phi, d=1.,
                                 epsv = (1,1,1), epsa = (0,0,0.), mode = +1, betamax = BETAMAX, out = None):
    dmat = E_diffraction_matrix(shape, ks, d, epsv, epsa, mode = mode, betamax = betamax)
    cmat = E_correction_matrix(beta, phi, ks, d, epsv, epsa, mode = mode)
    return dotmm(cmat,dmat, out = None)


@cached_function
def corrected_Epn_diffraction_matrix(shape, ks, beta,phi, d=1.,
                                 epsv = (1,1,1), epsa = (0,0,0.), betamax = BETAMAX, out = None):
    ep = corrected_E_diffraction_matrix(shape, ks, beta,phi, d= d,
                                 epsv = epsv, epsa = epsa, mode = +1, betamax = betamax)
    en = corrected_E_diffraction_matrix(shape, ks, beta,phi, d= d,
                                 epsv = epsv, epsa = epsa, mode = -1, betamax = betamax)
    return ep, en


@cached_function
def corrected_Epn_diffraction_matrix2(shape, ks, beta,phi, d=1.,
                                 epsv = (1,1,1), epsa = (0,0,0.), betamax = BETAMAX, out = None):
    ep = corrected_E_diffraction_matrix(shape, ks, beta,phi, d= d,
                                 epsv = epsv, epsa = epsa, mode = +1, betamax = betamax)
    en = corrected_E_diffraction_matrix(shape, ks, beta,phi, d= d,
                                 epsv = epsv, epsa = epsa, mode = -1, betamax = betamax)
    
    shape = ep.shape[:-2] + (4,4)
    out = np.zeros(shape = shape, dtype = ep.dtype)
    out[...,::2,::2] = ep
    out[...,1::2,1::2] = en
    return out

    
@cached_function
def corrected_field_diffraction_matrix(shape, ks, beta,phi, d=1.,
                                 epsv = (1,1,1), epsa = (0,0,0.), betamax = BETAMAX, out = None):
 
    dmat = field_diffraction_matrix(shape, ks, d, epsv, epsa, betamax = betamax)
    cmat = correction_matrix(beta, phi, ks, d/2, epsv, epsa)
    dmat = dotmm(dmat,cmat, out = out)
    return dotmm(cmat,dmat, out = dmat) 

@cached_function
def first_field_diffraction_matrix(shape, ks, beta,phi, d=1.,
                                 epsv = (1,1,1), epsa = (0,0,0.), betamax = BETAMAX, out = None):
    dmat = field_diffraction_matrix(shape, ks, d, epsv, epsa, betamax = betamax)
    cmat = field_correction_matrix(beta, phi, ks, d, epsv, epsa)
    return dotmm(dmat,cmat, out = None)

@cached_function
def second_field_diffraction_matrix(shape, ks, beta,phi, d=1.,
                                 epsv = (1,1,1), epsa = (0,0,0.), betamax = BETAMAX, out = None):
    dmat = field_diffraction_matrix(shape, ks, d, epsv, epsa, betamax = betamax)
    cmat = field_correction_matrix(beta, phi, ks, d, epsv, epsa)
    return dotmm(cmat,dmat, out = None)

@cached_function
def first_Epn_diffraction_matrix(shape, ks,  d = 1., epsv = (1,1,1), epsa = (0,0,0.),  betamax = BETAMAX, out = None):
    ks = np.asarray(ks, dtype = FDTYPE)
    epsv = np.asarray(epsv, dtype = CDTYPE)
    epsa = np.asarray(epsa, dtype = FDTYPE)
    alpha, f, fi= diffraction_alphaffi(shape, ks, epsv = epsv, epsa = epsa, betamax = betamax)
    kd = ks * d
    e = E_mat(f, mode = None)
    pmat = phase_mat(alpha, kd[...,None,None])
    return dotmdm(e,pmat,fi,out = out) 

@cached_function
def second_Epn_diffraction_matrix(shape, ks,  d = 1., epsv = (1,1,1), epsa = (0,0,0.),  betamax = BETAMAX, out = None):
    ks = np.asarray(ks, dtype = FDTYPE)
    epsv = np.asarray(epsv, dtype = CDTYPE)
    epsa = np.asarray(epsa, dtype = FDTYPE)
    alpha, f = diffraction_alphaf(shape, ks, epsv = epsv, epsa = epsa, betamax = betamax)
    kd = ks * d
    e = E_mat(f, mode = None)
    ei = inv(e)
    pmat = phase_mat(alpha, kd[...,None,None])
    return dotmdm(f,pmat,ei,out = out) 

@cached_function
def Epn_correction_matrix(beta,phi,ks, d=1., epsv = (1,1,1), epsa = (0,0,0.), out = None):
    alpha, f = alphaf(beta,phi,epsv,epsa)  
    kd = -np.asarray(ks)*d
    pmat = phase_mat(alpha, kd[...,None,None]) 
    e = E_mat(f, mode = None)
    ei = inv(e)
    return dotmdm(e,pmat,ei, out = out)

@cached_function
def first_corrected_Epn_diffraction_matrix(shape, ks, beta,phi, d=1.,
                                 epsv = (1,1,1), epsa = (0,0,0.), window = None, betamax = BETAMAX):
    
    dmat = first_Epn_diffraction_matrix(shape, ks, d, epsv, epsa, betamax = betamax)
    
    if window is not None:
        cmats = Epn_correction_matrix(beta, phi, ks, d, epsv, epsa)
        out = np.zeros_like(dmat)
        for w, cmat in zip(window,cmats):
            out += dotmm(cmat,w* dmat)
        return out
            
    else:
        
        cmat = Epn_correction_matrix(beta, phi, ks, d, epsv, epsa)
        return dotmm(cmat,dmat)

@cached_function
def second_corrected_Epn_diffraction_matrix(shape, ks, beta,phi, d=1.,
                                 epsv = (1,1,1), epsa = (0,0,0.), window = None, betamax = BETAMAX):
    dmat = second_Epn_diffraction_matrix(shape, ks, d, epsv, epsa, betamax = betamax)
    
    if window is not None:
        cmats = Epn_correction_matrix(beta, phi, ks, d, epsv, epsa)
        out = np.zeros_like(dmat)
        for w, cmat in zip(window,cmats):
            out += dotmm(w * dmat,cmat)
        return out
            
    else:
            
        cmat = Epn_correction_matrix(beta, phi, ks, d, epsv, epsa)
        return dotmm(dmat, cmat, out = None)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Window functions.
"""

from __future__ import absolute_import, print_function, division

import numpy as np
from dtmm.conf import FDTYPE, cached_function

def _r(shape, scale = 1.):
    """Returns radius array of a given shape."""
    ny,nx = [l/2 for l in shape]
    ay, ax = [np.arange(-l / 2. + .5, l / 2. + .5) for l in shape]
    xx, yy = np.meshgrid(ax, ay, indexing = "xy")
    r = ((xx/(nx*scale))**2 + (yy/(ny*scale))**2) ** 0.5    
    return r

def _r2(shape):
    """Returns radius array of a given shape."""
    ay, ax = [np.arange(-l / 2. + .5, l / 2. + .5) for l in shape]
    xx, yy = np.meshgrid(ax, ay, indexing = "xy")
    r = ((xx)**2 + (yy)**2) ** 0.5    
    return r

@cached_function
def blackman(shape, out = None):
    """Returns a blacman window of a given shape.
    
    Parameters
    ----------
    shape : (int,int)
        A shape of the 2D window.
    out : ndarray, optional
        Output array.
        
    Returns
    -------
    window : ndarray
        A Blackman window
    """
    r = _r(shape)  
    if out is None:
        out = np.ones(shape, FDTYPE)
    out[...] = 0.42 + 0.5*np.cos(1*np.pi*r)+0.08*np.cos(2*np.pi*r)
    mask = (r>= 1.)
    out[mask] = 0.
    return out

def gaussian(shape, waist):
    """Gaussian amplitude window function.
    
    Parameters
    ----------
    shape : (int,int)
        A shape of the 2D window
    waist : float
        Waist of the gaussian.
       
    Returns
    -------
    window : ndarray
        Gaussian beam window
    """
    r = _r(shape, waist)
    out = np.empty(shape, FDTYPE)
    return np.exp(-r**2, out = out)

def gaussian_beam(shape, waist, k0, z = 0., n = 1):
    """Returns gaussian beam window function. 
    
    Parameters
    ----------
    shape : (int,int)
        A shape of the 2D window
    waist : float
        Waist of the beam 
    k0 : float
        Wavenumber
    z : float, optional
        The z-position of waist (0. by default)
    n : float, optional
        Refractive index (1. by default)    
    out : ndarray, optional
        Output array.
       
    Returns
    -------
    window : ndarray
        Gaussian beam window
    """
    w0 = waist
    k = k0*n
    z0 = w0**2 *k/2.
    Rm2 = z/(z0**2+z**2)/2.
    w = w0 * (1+ (z/z0)**2)**0.5
    psi = np.arctan(z/z0)
    r = _r2(shape)
    return w0/w*np.exp(-(r/w)**2)*np.exp(1j*(k*z- psi))*np.exp(1j*(r*Rm2)**2)
    
def aperture(shape, diameter = 1., smooth = 0.05, out = None):
    """Returns aperture window function. 
    
    Aperture is basically a tukey filter with a given diameter
    
    Parameters
    ----------
    shape : (int,int)
        A shape of the 2D window
    diameter : float
        Width of the aperture (1. for max height/width)
    smooth : float
        Smoothnes parameter - defines smoothness of the edge of the aperture
        (should be between 0. and 1.; 0.05 by default)
    out : ndarray, optional
        Output array.
        
    Returns
    -------
    window : ndarray
        Aperture window
    """
    r = _r(shape, scale = diameter)
    return tukey(r,smooth, out = out)

def tukey(r,alpha = 0.1, rmax = 1., out =  None):
    if out is None:
        out = np.ones(r.shape, FDTYPE)
    else:
        out[...] = 1.
    r = np.asarray(r, FDTYPE)
    alpha = alpha * rmax
    mask = r > rmax -alpha
    if alpha > 0.:
        tmp = 1/2*(1-np.cos(np.pi*(r-rmax)/alpha))
        out[mask] = tmp[mask]
    mask = (r>= rmax)
    out[mask] = 0.
    return out  

__all__ = ["aperture", "blackman", "gaussian_beam", "gaussian"]

if __name__ == "__main__":
    import matplotlib.pyplot as plt
    plt.subplot(121)
    plt.imshow(aperture((32,32),diameter = 0.9))
    plt.subplot(122)
    plt.imshow(gaussian((32,32),0.5))


"""
Field denoising functions.
"""

import numpy as np
from dtmm.conf import FDTYPE
from dtmm.wave import betaphi
from dtmm.fft import fft2,ifft2

def tukey_notch_filter(x,x0,alpha):
    x = np.asarray(x, FDTYPE)
    out = np.ones(x.shape, FDTYPE)
    alpha = alpha * x0
    mask = (x < x0 + alpha)
    mask = np.logical_and(mask, (x > x0 - alpha))
    if alpha > 0.:
        tmp = 1/2*(1-np.cos(np.pi*(x-x0)/alpha))
        out[mask] = tmp[mask]
    return out  

def exp_notch_filter(x,x0,sigma):
    return np.asarray((1 - 1*np.exp(-np.abs(x-x0).clip(0,x0)/sigma))/(1-1*np.exp(-x0/sigma)),FDTYPE)

def denoise_field(field, wavenumbers, beta , smooth = 1, filter_func = exp_notch_filter, out = None):
    """Denoises field by attenuating modes around the selected beta parameter.
    """
    ffield = fft2(field, out = out)
    ffield = denoise_fftfield(ffield, wavenumbers, beta, smooth = smooth, filter_func = filter_func, out = ffield)
    return ifft2(ffield, out = ffield)

def denoise_fftfield(ffield, wavenumbers, beta, smooth = 1, filter_func = exp_notch_filter, out = None):
    """Denoises fourier transformed field by attenuating modes around the selected beta parameter.
    """
    shape = ffield.shape[-2:]
    b, p = betaphi(shape,wavenumbers)
    f = filter_func(b, beta, smooth)
    f = f[...,None,:,:]
    ffield = np.multiply(ffield, f, out = out)
    return ffield



# -*- coding: utf-8 -*-
"""
Color conversion functions and utilities.

CMF creation function
---------------------

* :func:`.load_cmf` : loads specter cmf data from file or pre-defined table
* :func:`.load_tcmf` : loads transmission cmf data from file or pre-defined table
* :func:`.cmf2tcmf` : converts cmf to transmission cmf.
* :func:`.srf2cmf` : converts spectral respone data to cmf
* :func:`.load_specter` : load specter from file or from data.
* :func:`.normalize_specter` : for specter normalization.

Color conversion
----------------

* :func:`.specter2color` : converts specter data to color RGB or gray image.
* :func:`.apply_gamma` : applies gamma curve to linear data
* :func:`.apply_srgb_gamma` : applies sRGB gamma curve to linear data
* :func:`.xyz2rgb` : Converts XYZ data to RGB
* :func:`.xyz2gray` : Converts XYZ data to YYY (gray)
* :func:`.spec2xyz` : Converts specter to XYZ

"""

from __future__ import absolute_import, print_function, division

from dtmm.conf import FDTYPE, NUMBA_TARGET, NFDTYPE, NUMBA_CACHE, DATAPATH, CMF

import numpy as np
import numba
import os

#DATAPATH = os.path.join(os.path.dirname(__file__), "data")


# D65 standard light 5nm specter
D65PATH = os.path.join(DATAPATH, "D65.dat" )
#: color matrix for sRGB color space in D65 reference white
XYZ2RGBD65 = np.array([[ 3.2404542, -1.5371385, -0.4985314],
                       [-0.9692660,  1.8760108,  0.0415560],
                       [ 0.0556434, -0.2040259,  1.0572252]])
RGB2XYZ = np.linalg.inv(XYZ2RGBD65)
#Srgb tranfer function constants
SRGBIGAMMA = 1/2.4
SRGBSLOPE = 12.92
SRGBLINPOINT = 0.0031308
SRGBA = 0.055

@numba.vectorize([NFDTYPE(NFDTYPE, NFDTYPE)], nopython = True, target = NUMBA_TARGET, cache = NUMBA_CACHE) 
def apply_gamma(value, gamma):
    """apply_gamma(value, gamma)
    
Applies standard gamma function (transfer function) to the given (linear) data.
    
Parameters
----------
value : float
    Input value
gamma : float
    Gamma factor"""
    if value > 1.:
        return 1.
    if value < 0:
        return 0.
    else:
        return value**(1./gamma)

@numba.vectorize([NFDTYPE(NFDTYPE)], nopython = True, target = NUMBA_TARGET, cache = NUMBA_CACHE) 
def apply_srgb_gamma(value):
    """apply_srgb_gamma(value)

Applies sRGB gamma function (transfer function) to the given (linear) data."""
    if value < SRGBLINPOINT:
        if value < 0.:
            return 0.
        else:
            return SRGBSLOPE * value
    else:
        if value > 1.:
            return 1.
        else:
            return (1+SRGBA)*value**SRGBIGAMMA-SRGBA

@numba.guvectorize([(NFDTYPE[:], NFDTYPE[:])], '(n)->(n)', target = NUMBA_TARGET, cache = NUMBA_CACHE)
def xyz2srgb(xyz, rgb):
    """xyz2srgb(xyz)
    
Converts XYZ value to RGB value based on the sRGB working space with 
D65 reference white.
"""
    assert len(xyz) >= 3
    xyz0 = xyz[0]
    xyz1 = xyz[1]
    xyz2 = xyz[2]
    for k in range(3):
        rgb[k] = XYZ2RGBD65[k,0] * xyz0 +  XYZ2RGBD65[k,1]* xyz1 +  XYZ2RGBD65[k,2]* xyz2
        
@numba.guvectorize([(NFDTYPE[:], NFDTYPE[:])], '(n)->(n)', target = NUMBA_TARGET, cache = NUMBA_CACHE)
def xyz2gray(xyz, gray):
    """xyz2gray(xyz)
    
Converts XYZ value to Gray color"""
    assert len(xyz) >= 3
    y = xyz[1]
    for k in range(3):
        gray[k] = y
                
@numba.guvectorize([(NFDTYPE[:],NFDTYPE[:,:],NFDTYPE[:])], '(n),(n,m)->(m)', target = NUMBA_TARGET, cache = NUMBA_CACHE)
def spec2xyz(spec,cmf,xyz):
    """spec2xyz(spec,cmf)
    
Converts specter array to xyz value.
    
Parameters
----------
spec : array_like
    Input specter data
cmf : array_like
    Color matching function
    
Returns
-------
xyz : ndarray
    Computed xyz value."""  
    
    for j in range(cmf.shape[1]):
        xyz[j] = 0.
        for i in range(cmf.shape[0]):
            xyz[j] = xyz[j] + cmf[i,j]*spec[i]          
            
def specter2color(spec, cmf, norm = False, gamma = True, gray = False, out = None):
    """Converts specter data to RGB data (color or gray).
    
    Specter shape must be [...,k], where wavelengths are in the last axis. cmf 
    must be a valid color matchin function array of size [k,3].
    
    Parameters
    ----------
    spec : array
        Specter data of shape [..., n] where each data element in the array has
        n wavelength values
    cmf : array
        A color matching function (array of shape [n,3]) that converts the specter data 
        to a XYZ color.
    norm : bool or float, optional
        If set to False, no data normalization is performed (default). If True,
        internally, xyz data is normalized in the range [0,1.], so that no clipping occurs.
        If it is a float, data is normalized to this value.
    gamma : bool or float, optional
        If gamma is True srgb gamma function is applied (default). If float is
        provided, standard gamma factor is applied with a given gamma value. If False,
        no gamma correction is performed.
    gray : bool, optional
        Whether gray output is calculated (color by default)
    out : array, optional
        Output array of shape (...,3)
        
    Returns
    -------
    rgb : ndarray
        A computed RGB value.
        
    Notes
    -----
    Numpy broadcasting rules apply to spec and cmf.
        
    Example
    -------
    >>> cmf = load_tcmf()
    >>> specter2color([1]*81, cmf)#should be close to 1,1,1
    ... # doctest: +NORMALIZE_WHITESPACE
    array([0.99994901, 1.        , 0.99998533])
    """
    #if isinstance(spec, list):
    #    spec = np.add.reduce(spec)
    
    cmf = np.asarray(cmf)
    if cmf.shape[-1] != 3:
        raise ValueError("Grayscale cmf! Cannot convert to color.")
      
    out = spec2xyz(spec,cmf, out)
        
    if norm is True:
        #normalize to max in any of the XYZ channels.. so that no clipping occurs.
        out = np.divide(out,out.max(),out)
        
    elif norm != 0:
        out = np.divide(out,norm,out)
 
    if gray == True:
        out = xyz2gray(out,out) 
    else:
        out = xyz2srgb(out,out)
    if gamma is True:
        apply_srgb_gamma(out,out)
    elif gamma is not False:
        apply_gamma(out,gamma,out)
    
    return out

def srf2cmf(srf, out = None):
    """Converts spectral response function (Y) to color matching function (XYZ).
    
    Parameters
    ----------
    srff : array_like
        Spectral response function of shape (...,n)
    out : ndarray, optional
        Output array.
        
    Returns
    -------
    cmf: ndarray
        A color matching function array of shape (...,n,3)
    """
    if out is None:
        out = np.empty(shape = srf.shape + (3,), dtype = srf.dtype)
    
    out[...,0] = RGB2XYZ[0,0] * srf +  RGB2XYZ[0,1] * srf +  RGB2XYZ[0,2] * srf
    out[...,1] = srf
    out[...,2] = RGB2XYZ[2,0] * srf +  RGB2XYZ[2,1] * srf +  RGB2XYZ[2,2] * srf
    return out
    
def normalize_specter(spec, cmf, out = None):
    """Normalizes specter based on the color matching function. (cmf array) so that
    calculated Y value is 1.
    
    Parameters
    ----------
    spec : array_like
        Input illuminant specter data of shape (...,n).
    cmf : array_like
        Color matching function of shape (...,n,3).
    out : ndarray, optional
        Output array.
        
    Returns
    -------
    normalized_spec : ndarray
        A normalized version of the input specter
        
    Notes
    -----
    Numpy broadcasting rules apply to spec and cmf.
    """
    cmf = np.asarray(cmf)
    if cmf.shape[-1] == 3:
        #cmf is color matching function
        xyz = spec2xyz(spec,cmf)
        norm = xyz[...,1] #Y value is ligtness.. normalize it to this
    else:
        raise ValueError("Incompatible cmf shape")
    return np.divide(spec,norm,out)

def load_tcmf(wavelengths = None, illuminant = "D65", cmf = CMF, norm = True, retx = False, 
              single_wavelength = False):
    """Loads transmission color matching function.
    
    This functions loads a CIE XYZ color matching function and transforms it
    to a transmission color matching function for a given illuminant. Resulting 
    CMF matrix will transform unity into white color.
    
    Parameters
    ----------
    wavelengths : array_like, optional
        Wavelengths at which data is computed. If not specified (default), original
        data from the 5nm tabulated data is returned.
    illuminant : str, optional
        Name of the standard illuminant or path to illuminant data.
    cmf : str, optional
        Name or path to the cmf function. Can be 'CIE1931' for CIE 1931 2-deg 
        5nm tabulated data, 'CIE1964' for CIE1964 10-deg 5nm tabulatd data, or
        'CIE2006-2' or 'CIE2006-10' for a proposed CIE 2006 2- or 10-deg 5nm 
        tabulated data. 
    norm : int, optional
        By default cmf is normalized so that unity transmission value over the
        full spectral range of the illuminant is converted to XYZ color with Y=1.
        You can disable this by setting norm = 0. If you set norm = 2, then the
        cmf is normalized for the interpolated spectra at given wavelengths, 
        and not to the full bandwidth of the spectra (norm = 1).
    retx : bool, optional
        Should the selected wavelengths be returned as well.
    single_wavelength : bool, optional
        If specified, color matching function for single wavelengths specter is
        calculated by interpolation. By default, specter is assumed to be a 
        piece-wise linear function and continuous between the specified 
        wavelengts, and data is integrated instead.
    
    Returns
    -------
    cmf : array
        Color matching function array of shape [n,3] or a tuple of (x,cmf) 
        if retx is specified.
    
    
    Example
    -------
    >>> cmf = load_tcmf()
    >>> specter2color([1]*81, cmf) #should be close to 1,1,1
    ... # doctest: +NORMALIZE_WHITESPACE
    array([0.99994901, 1.        , 0.99998533])
    """
    if wavelengths is not None and len(wavelengths) == 1:
        single_wavelength = True

    if single_wavelength == True:
        x, cmf = load_cmf(wavelengths, single_wavelength = True,retx = True, cmf = cmf)
        spec = load_specter(wavelengths = x, illuminant = illuminant)
        cmf = cmf2tcmf(cmf, spec, norm = bool(norm))  
    else:
        x,cmf = load_cmf(retx = True, cmf = cmf)
        spec = load_specter(wavelengths = x, illuminant = illuminant)
        cmf = cmf2tcmf(cmf, spec, norm = bool(norm)) 
        if wavelengths is not None:
             cmf = integrate_data(wavelengths, x,cmf)
             x = wavelengths
             if norm == 2:
                 cmf = cmf2tcmf(cmf, [1.]*len(wavelengths), norm = norm)  
    if retx == True:
        return x, cmf
    else:
        return cmf

def cmf2tcmf(cmf, spec, norm = True, out = None):
    """Converts CMF table to specter-normalized transmission CMF table
    
    Parameters
    ----------
    cmf : array_like
        Color matchinf function array
    spec : array_like
        Illuminant specter array
    norm : bool, optional
       Whether to normalize illuminant specter before constructing the CMF.
    out : ndarray, optional
       Output array
    
    Returns
    -------
    out : ndarray
        A transmission color matching function array.
       
    Notes
    -----
    Numpy broadcasting rules apply to spec and cmf.
    """
    cmf = np.asarray(cmf,FDTYPE)
    spec = np.asarray(spec,FDTYPE)
    if norm == True:
        spec = normalize_specter(spec, cmf)
    return np.multiply(spec[:,np.newaxis],cmf, out = out) 
    
def load_specter(wavelengths = None, illuminant = "D65", retx = False):
    """Loads illuminant specter data from file.
    
    Parameters
    ----------
    wavelengths : array_like, optional
        Wavelengths at which data is interpolated
    illuminant : str, or array, optional
        Name of the standard illuminant or filename. If specified as array, it must
        be an array of shape (n,2). The first column decribes wavelength and the
        second is the intensity.
    retx : bool, optional
        Should the selected wavelengths be returned as well.
        
    Returns
    -------
    specter : array
        Specter array of shape [num] or a tuple of (x,specter) 
        if retx is specified
        
    Examples
    --------
    
    #D65 evaluated at three wavelengths
    >>> spec = load_specter((450,500,550), "D65") 
    
    #illuminant with triangular specter evaluated at three wavelengths
    >>> spec = load_specter([450,475,500,], illuminant = [[400,0],[500,1],[600,0]]) 
    
    """   
    if isinstance(illuminant, str):    
        try:
            # predefined data in a file
            data = np.loadtxt(os.path.join(DATAPATH, illuminant + ".dat"))
        except:
            data = np.loadtxt(illuminant)
    else:
        data = np.asarray(illuminant)
        if data.ndim != 2 and data.shape[-1] != 2:
            raise ValueError("Not a valid illuminant data")
        
    if wavelengths is not None:
        data = interpolate_data(wavelengths, data[:,0], data[:,1:])
        data = np.ascontiguousarray(data[:,0], dtype = FDTYPE)
    else:
        wavelengths = np.ascontiguousarray(data[:,0], dtype = FDTYPE)
        data = np.ascontiguousarray(data[:,1], dtype = FDTYPE)
        
    if retx == True:
        return wavelengths, data
    else:
        return data
    
def load_cmf(wavelengths = None, cmf = CMF, retx = False, single_wavelength = False):
    """Load XYZ Color Matching function.
    
    This function loads tabulated data and re-calculates xyz array on a 
    given range of wavelength values.
    
    See also load_tcmf.
    
    Parameters
    ----------
    wavelengths : array_like, optional
        A 1D array of wavelengths at which data is computed. If not specified 
        (default), original data from the 5nm tabulated data is returned.
    cmf : str, optional
        Name or path to the cmf function. Can be 'CIE1931' for CIE 1931 2-deg 
        5nm tabulated data, 'CIE1964' for CIE1964 10-deg 5nm tabulated data, or
        'CIE2006-2' or 'CIE2006-10' for a proposed CIE 2006 2- or 10-deg 5nm 
        tabulated data. For grayscale cameras, there is a 'CMOS' spectral 
        response data. You can also provide 'FLAT' for flat (unity) response function.
    retx : bool, optional
        Should the selected wavelengths be returned as well.
    single_wavelength : bool, optional
        If specified, color matching function for single wavelengths specter is
        calculated by interpolation. By default, specter is assumed to be a 
        piece-wise linear function and continuous between the specified 
        wavelengts, and data is integrated instead.
   
    Returns
    -------
    cmf : array
        Color matching function array of shape [n,3] or a tuple of (x,cmf) 
        if retx is specified.
    """
    try:
        if cmf == "FLAT":
            if wavelengths is None:
                wavelengths = np.arange(380,781,5)
            data = np.zeros((len(wavelengths),3))
            data[:,1] = 100. #100% QE
            if retx == True:
                return wavelengths, data
            else:
                return data       
        
        if cmf.startswith("CIE"):
            data = np.loadtxt(os.path.join(DATAPATH, cmf + "XYZ.dat"))
        else:
            data = np.loadtxt(os.path.join(DATAPATH, cmf + "Y.dat"))
    except:
        data = np.loadtxt(cmf)
    
    if data.shape[-1] == 4:
        x, data = np.ascontiguousarray(data[:,0], dtype = FDTYPE),  np.ascontiguousarray(data[:,1:], dtype = FDTYPE)
    elif data.shape[-1] == 2:
        x, data = np.ascontiguousarray(data[:,0], dtype = FDTYPE),  np.ascontiguousarray(data[:,1], dtype = FDTYPE)
    else:
        raise ValueError("Not a valid cmf data!")
        
    if wavelengths is not None:
        wavelengths = np.asarray(wavelengths, dtype = FDTYPE)
        if wavelengths.ndim != 1:
            raise ValueError("Wavelengths has to be 1D array")
        if len(wavelengths) == 1:
            single_wavelength = True
    if single_wavelength == True and wavelengths is not None:
        data = interpolate_data(wavelengths, x, data)
        x = wavelengths
    elif wavelengths is not None:
        data = integrate_data(wavelengths, x,data)
        x = wavelengths

    if data.ndim == 1:
        #convert spectral response to cmf
        data = srf2cmf(data)
    if retx == True:
        return x, data
    else:
        return data
    
#import scipy.interpolate as interpolate

def interpolate_data(x, x0, data):
    data = np.asarray(data, dtype = FDTYPE)
    x0 = np.asarray(x0)
    x = np.asarray(x)
    if data.ndim in (1,2) and x0.ndim == 1 and x.ndim == 1: 
        if data.ndim == 2:
            out = np.zeros(shape = x.shape + data.shape[1:], dtype = data.dtype)  
            rows, cols = data.shape  
            for i in range(cols):
                #f = interpolate.interp1d(x0, data[:,i],fill_value = 0, kind="linear")
                #out[...,i] = f(x)
                out[...,i] = np.interp(x, x0, data[:,i],left = 0., right = 0.)
            return out 
        else:
            return np.interp(x, x0, data, left = 0., right = 0.)
    else:
        raise ValueError("Invalid dimensions of input data.")
              

#def integrate_data(x,x0,cmf):
#    cmf = np.asarray(cmf)
#    x0 = np.asarray(x0)
#    x = np.asarray(x)
#    out = np.zeros(shape = (len(x), cmf.shape[1]), dtype = cmf.dtype) 
#    n = len(x)
#    triang = np.array([0.,1,0])
#    for i in range(n):
#        if i == 0:
#            data = np.interp(x0,x[0:i+2],triang[1:],left = 0.)
#        elif i == n-1:
#            data = np.interp(x0,x[i-1:i+1],triang[:-1],right = 0.)
#        else:
#            data = np.interp(x0,x[i-1:i+2],triang)      
#        out[i,:] = (cmf*data[:,np.newaxis]).sum(0)
#    return out
 
def integrate_data(x,x0,cmf):
    cmf = np.asarray(cmf)
    x0 = np.asarray(x0)
    xout = np.asarray(x)
    ndim = cmf.ndim
    if ndim in (1,2) and x0.ndim == 1 and xout.ndim == 1:    
        dx = x0[1]-x0[0]
        out = np.zeros(shape = (len(x),)+cmf.shape[1:], dtype = cmf.dtype) 
        n = len(x)
        for i in range(n):
            if i == 0:
                x,y = _rxn(xout,i,dx,ndim)
                
                data = (interpolate_data(x,x0,cmf)*y).sum(0)
            elif i == n-1:
                x,y  = _lxn(xout,i,dx,ndim)
                data = (interpolate_data(x,x0,cmf)*y).sum(0) 
            else:
                x,y  = _rxn(xout,i,dx,ndim)
                tmp = (interpolate_data(x,x0,cmf)*y)
                tmp[0] = tmp[0]/2 #first element is counted two times...
                data = tmp.sum(0)
                x,y  = _lxn(xout,i,dx,ndim)
                tmp = (interpolate_data(x,x0,cmf)*y)
                tmp[0] = tmp[0]/2 #first element is counted two times...
                data += tmp.sum(0)
            out[i,...] = data
        return out 

            
    else:
        raise ValueError("Invalid dimensions of input data.")
 
def _rxn(x,i,dx,ndim):
    xlow, xhigh = x[i], x[i+1]
    dx = (xhigh - xlow)/dx
    if dx < 1:
        import warnings
        warnings.warn("The resolution of the integrand is too low.", stacklevel = 2)
    #n = int(round(dx))+1
    n = int(dx-1)+1
    dx = dx/n
    xout = np.linspace(xlow,xhigh,n+1)
    yout = np.linspace(1.,0.,n+1)*dx
    if ndim == 2:
        return xout,yout[:,None]
    else:
        return xout,yout

def _lxn(x,i,dx, ndim):
    xlow, xhigh = x[i-1], x[i]
    dx = (xhigh - xlow)/dx
    if dx < 1:
        import warnings
        warnings.warn("The resolution of the integrand is too low.", stacklevel = 2)
    #n = int(round(dx))+1
    n = int(dx-1)+1
    dx = dx/n
    xout = np.linspace(xhigh,xlow,n+1)
    yout = np.linspace(1.,0.,n+1)*dx
    if ndim == 2:
        return xout,yout[:,None]
    else:
        return xout,yout


#__all__ = ["specter2color", "load_tcmf", "load_cmf", "load_specter"]
                    
   
if __name__ == "__main__":
    import doctest
    doctest.testmod()
    import matplotlib.pyplot as plt
    num = 81
    for num in (81,41,21,11,5):
        wavelengths = np.linspace(380,780,num)
        x,xyz = load_tcmf(wavelengths, norm = True,retx = True, single_wavelength = True)
        
        plt.plot(x,xyz[:,0],label = num) #X color
        x,xyz = load_tcmf(wavelengths, norm = True,retx = True, single_wavelength = False)
        
        plt.plot(x,xyz[:,0],"o") #X color
    plt.legend()
    plt.show()

    



"""
Configuration and constants
===========================

dtmm configuration functions and constants

"""

from __future__ import absolute_import, print_function, division

import numpy as np
from functools import wraps
import os, warnings, shutil

try:
    from configparser import ConfigParser
except:
    #python 2.7
    from ConfigParser import ConfigParser
    
    
#These will be defined later at runtime... here we hold reference to disable warnings in autoapi generation
I = None
CDTYPE = None 
IDTYPE = None
C = None
NUMBA_PARALLEL = None
NCDTYPE = None
NFDTYPE = None


DATAPATH = os.path.dirname(__file__)

def read_environ_variable(name, default):
    try:
        return int(os.environ.get(name, default))
    except ValueError:
        return int(default)
        warnings.warn("Environment variable {0:s} was found, but its value is not valid!".format(name))

def get_home_dir():
    """
    Return user home directory
    """
    try:
        path = os.path.expanduser('~')
    except:
        path = ''
    for env_var in ('HOME', 'USERPROFILE', 'TMP'):
        if os.path.isdir(path):
            break
        path = os.environ.get(env_var, '')
    if path:
        return path
    else:
        raise RuntimeError('Please define environment variable $HOME')

#: home directory
HOMEDIR = get_home_dir()

DTMM_CONFIG_DIR = os.path.join(HOMEDIR, ".dtmm")

NUMBA_CACHE_DIR = os.path.join(DTMM_CONFIG_DIR, "numba_cache")


if not os.path.exists(DTMM_CONFIG_DIR):
    try:
        os.makedirs(DTMM_CONFIG_DIR)
    except:
        warnings.warn("Could not create folder in user's home directory! Is it writeable?",stacklevel=2)
        NUMBA_CACHE_DIR = ""

#FILE_LOCK = os.path.join(DTMM_CONFIG_DIR, "lock")        
# if os.path.exists(NUMBA_CACHE_DIR):
#     #we have compiled functions.. make sure it is safe to read from this cache folder
#     if os.path.exists(FILE_LOCK):
#         #it is not safe
#         warnings.warn("There appears to be another instance of dtmm running, so caching is disabled. Try removing .dtmm/numba_cache folder if this message persists.",stacklevel=2)
#         NUMBA_CACHE_DIR = ""  
#     else:
#         #it is safe, make a file lock

#         f = open(FILE_LOCK, "w")
#         f.close()

# import atexit

# @atexit.register
# def cleanup():
#     if NUMBA_CACHE_DIR != "":
#         os.remove(FILE_LOCK)      

CONF = os.path.join(DTMM_CONFIG_DIR, "dtmm.ini")
CONF_TEMPLATE = os.path.join(DATAPATH, "dtmm.ini")

config = ConfigParser()

if not os.path.exists(CONF):
    try:
        shutil.copy(CONF_TEMPLATE, CONF)
    except:
        warnings.warn("Could not copy config file in user's home directory! Is it writeable?")
        CONF = CONF_TEMPLATE
config.read(CONF)
    
def _readconfig(func, section, name, default):
    try:
        return func(section, name)
    except:
        #warnings.warn("There was a problem parsing config file while reading section {} and option {}; using default value instead".format(section, name))
        return default
    
    
def is_module_installed(name):
    """Checks whether module with name 'name' is istalled or not"""
    try:
        __import__(name)
        return True
    except ImportError:
        return False    
        
NUMBA_INSTALLED = is_module_installed("numba")
MKL_FFT_INSTALLED = is_module_installed("mkl_fft")
SCIPY_INSTALLED = is_module_installed("scipy")

BETAMAX = _readconfig(config.getfloat, "core", "betamax", 0.8)
SMOOTH = _readconfig(config.getfloat, "core", "smooth", 0.1)


#setting environment variables does not seem to work properly in numba...
#disable cache dir until I figure out how to do it properly.
    
# if NUMBA_CACHE_DIR != "":
#     os.environ["NUMBA_CACHE_DIR"] = NUMBA_CACHE_DIR #set it to os.environ.. so that numba can use it

# def wait_until_cache_is_set(timeout = 1):
#     import subprocess, time
#     t = time.time()
#     while True:
#         print()
#         out = subprocess.run(["printenv"], capture_output=True, text = True)
#         if out.find("NUMBA_CACHE_DIR") != -1:
#             break
#         if time.time()-t > timeout:
#             NUMBA_CACHE_DIR = ""


if read_environ_variable("DTMM_TARGET_PARALLEL",
            default = _readconfig(config.getboolean, "numba", "parallel", False)):
    NUMBA_TARGET = "parallel"
    NUMBA_PARALLEL = True
else:
    NUMBA_TARGET = "cpu"
    NUMBA_PARALLEL = False

NUMBA_CACHE = False   

_numba_0_39_or_greater = False
_numba_0_45_or_greater = False
try: 
    import numba as nb
    major, minor = nb.__version__.split(".")[0:2]
    if int(major) == 0 and int(minor) >=45:
        _numba_0_45_or_greater = True
    if int(major) == 0 and int(minor) >=39:
        _numba_0_39_or_greater = True
except:
    print("Could not determine numba version you are using, assuming < 0.39")


if read_environ_variable("DTMM_NUMBA_CACHE",
            default = _readconfig(config.getboolean, "numba", "cache", True)):
    if NUMBA_CACHE_DIR == "":
        NUMBA_CACHE = False
    elif NUMBA_PARALLEL == False:
        NUMBA_CACHE = True  
    else:
        try: 
            import numba as nb
            major, minor = nb.__version__.split(".")[0:2]
            if int(major) == 0 and int(minor) >=45:
                NUMBA_CACHE = True 
            else:
                NUMBA_CACHE = False
                warnings.warn("Numba caching was disabled because version of numba < 0.45 cannot use caching with parallelization enabled!")
        except:
            NUMBA_CACHE = False

if read_environ_variable("DTMM_FASTMATH",
        default = _readconfig(config.getboolean, "numba", "fastmath", False)):        
    NUMBA_FASTMATH = True
else:
    NUMBA_FASTMATH = False



#reference to all cashed functions - for automatic cache clearing with clear_cache.
_cache = set()

def clear_cache(func = None):
    """Clears compute cache.
    
    Parameters
    ----------
    func : function
        A cached function of which cache results are to be cleared (removed
        from cache). If not provided (default) all cache data is cleared from 
        all registred cached function - functions returned by the 
        :func:`cached_function` decorator"""
    
    if func is not None:
        func.cache.clear()
    else:
        for func in _cache:
            func.cache.clear()
        
            
def cached_function(f):
    """A decorator that converts a function into a cached function. 
    
    The function needs to be a function that returns a numpy array as a result.
    This result is then cached and future function calls with same arguments
    return result from the cache. Function arguments must all be hashable, or
    are small numpy arrays. The function can also take "out" keyword argument for
    an output array in which the resulting array is copied to.
    
    Notes
    -----
    When caching is enabled, cached numpy arrayes have a read-only attribute. 
    You need to copy first, or provide an output array if you need to write to 
    the result.
    """
    
    def pop_fifo_result_from_cache(cache):
        try:
            return cache.pop(next(iter(cache.keys())))
        except StopIteration:
            return None   
        
    def add_result_to_cache(result, key, cache):
        pop_fifo_result_from_cache(cache)   
        cache[key] = result

    def to_key(arg, name = None):
        from dtmm.hashing import hash_buffer 
        if isinstance(arg, np.ndarray):
            arg = (arg.shape, arg.dtype, arg.strides, hash_buffer(arg))
            #return (arg.shape, arg.dtype, tuple(arg.flat))
        if name is None:
            return arg
        else:
            return (name, arg)
        
    def copy(result,out):
        if out is not None:
            if isinstance(result, tuple):
                for o,a in zip(out,result):
                    o[...] = a
            else:
                out[...] = result
            return out
        else:
            return result   
        
    def set_readonly(result):
        if isinstance(result, tuple):
            for a in result:
                a.setflags(write = False)
        else:
            result.setflags(write = False)     
 
    def unset_readonly(result):
        if isinstance(result, tuple):
            for a in result:
                a.setflags(write = True)
        else:
            result.setflags(write = True)     
           
#    def delete(f):
#        f.cache.clear()
#        _cache.remove(f)
#        f.cache = None
   
     
    @wraps(f)
    def _f(*args,**kwargs):
        try_read_from_cache = (kwargs.pop("cache",True) == True) and (DTMMConfig.cache != 0) and _f.cache is not None
        if try_read_from_cache:
            key = tuple((to_key(arg) for arg in args)) + tuple((to_key(arg, name = key) for key,arg in kwargs.items()))
            out = kwargs.pop("out",None)    
            try:
                result = _f.cache[key]
                return copy(result,out)
            except KeyError:
                if kwargs.pop("reuse",False):
                    result = pop_fifo_result_from_cache(_f.cache)
                    unset_readonly(result)
                    kwargs["out"] = result 
                result = f(*args,**kwargs)
                set_readonly(result)
                add_result_to_cache(result,key, _f.cache)
                return copy(result,out)
        else:
            return f(*args,**kwargs)
    _f.cache = {}
    
    _cache.add(_f)
    #_f.delete = delete
    return _f    
 
def cached_result(f):
    """A decorator that converts a function into a cached result function. 
    
    The function needs to be a function that returns any result.
    Function arguments must all be hashable, or
    are small numpy arrays. 
    """
    
    def pop_fifo_result_from_cache(cache):
        try:
            return cache.pop(next(iter(cache.keys())))
        except StopIteration:
            return None   
        
    def add_result_to_cache(result, key, cache):
        pop_fifo_result_from_cache(cache)   
        cache[key] = result

    def to_key(arg, name = None):
        from dtmm.hashing import hash_buffer 
        if isinstance(arg, np.ndarray):
            arg = (arg.shape, arg.dtype, arg.strides, hash_buffer(arg))
            #return (arg.shape, arg.dtype, tuple(arg.flat))
        if name is None:
            return arg
        else:
            return (name, arg)

    def set_readonly(result):
        if isinstance(result, tuple):
            for a in result:
                try:
                    a.setflags(write = False)
                except AttributeError:
                    pass
        else:
            try:
                result.setflags(write = False)  
            except AttributeError:
                pass
            
    @wraps(f)
    def _f(*args,**kwargs):
        try_read_from_cache = (kwargs.pop("cache",True) == True) and (DTMMConfig.cache != 0) and _f.cache is not None
        if try_read_from_cache:
            key = tuple((to_key(arg) for arg in args)) + tuple((to_key(arg, name = key) for key,arg in kwargs.items()))  
            try:
                result = _f.cache[key]
                return result
            except KeyError:
                result = f(*args,**kwargs)
                set_readonly(result)
                add_result_to_cache(result,key, _f.cache)
                return result
        else:
            return f(*args,**kwargs)
    _f.cache = {}
    
    _cache.add(_f)
    #_f.delete = delete
    return _f      


def detect_number_of_cores():
    """
    detect_number_of_cores()

    Detect the number of cores in this system.

    Returns
    -------
    out : int
        The number of cores in this system.

    """
    import os
    # Linux, Unix and MacOS:
    if hasattr(os, "sysconf"):
        if "SC_NPROCESSORS_ONLN" in os.sysconf_names:
            # Linux & Unix:
            ncpus = os.sysconf("SC_NPROCESSORS_ONLN")
            if isinstance(ncpus, int) and ncpus > 0:
                return ncpus
        else:  # OSX:
            return int(os.popen2("sysctl -n hw.ncpu")[1].read())
    # Windows:
    if "NUMBER_OF_PROCESSORS" in os.environ:
        ncpus = int(os.environ["NUMBER_OF_PROCESSORS"])
        if ncpus > 0:
            return ncpus
    return 1  # Default

F32DTYPE = np.dtype("float32")
F64DTYPE = np.dtype("float64")
C64DTYPE = np.dtype("complex64")
C128DTYPE = np.dtype("complex128")
U32DTYPE = np.dtype("uint32")


if read_environ_variable("DTMM_DOUBLE_PRECISION",
  default =  _readconfig(config.getboolean, "core", "double_precision", True)):
    FDTYPE = F64DTYPE
    CDTYPE = C128DTYPE
    UDTYPE = U32DTYPE
    PRECISION = "double"
else:
    FDTYPE = F32DTYPE
    CDTYPE = C64DTYPE
    UDTYPE = U32DTYPE
    PRECISION = "single"
    
def disable_mkl_threading():
    """Disables mkl threading."""
    try:
        import mkl
        mkl.set_num_threads(1)
    except ImportError:
        warnings.warn("Cannot disable mkl threading, because 'mkl' module is not installed!")

def enable_mkl_threading():
    """Enables mkl threading."""
    try:
        import mkl
        mkl.set_num_threads(detect_number_of_cores())
    except ImportError:
        warnings.warn("Cannot enable mkl threading, because 'mkl' module is not installed!")

class DTMMConfig(object):
    """DTMM settings are here. You should use the set_* functions in the
    conf.py module to set these values"""
    def __init__(self):
        
        
        if MKL_FFT_INSTALLED:
            self.fftlib = "mkl_fft"
        elif SCIPY_INSTALLED:
            self.fftlib = "scipy"
        else:
            self.fftlib = "numpy"
        
        if _readconfig(config.getboolean, "fft", "parallel", False):
            self.nthreads = _readconfig(config.getint, "fft", "nthreads", 
                                        detect_number_of_cores())
        else:
            self.nthreads = 1
        if _readconfig(config.getboolean, "core", "cache", True):
            self.cache = 1
        else:
            self.cache = 0
        self.verbose = 0
        
        self.gray =  _readconfig(config.getboolean, "viewer", "gray", False)
        self.show_ticks = _readconfig(config.getboolean, "viewer", "show_ticks", None)
        self.show_scalebar = _readconfig(config.getboolean, "viewer", "show_scalebar", False)
        self.show_sliders = _readconfig(config.getboolean, "viewer", "show_sliders", True)
        gamma = _readconfig(config.getfloat, "viewer", "gamma", None)
        if gamma is None:
            gamma = _readconfig(config.getboolean, "viewer", "gamma", True)
        self.gamma = gamma
        
        self.n_cover = _readconfig(config.getfloat, "viewer", "n_cover", 1.5)
        self.d_cover = _readconfig(config.getfloat, "viewer", "d_cover", 0.)
        self.immersion = _readconfig(config.getboolean, "viewer", "immersion", False)
        self.NA = _readconfig(config.getfloat, "viewer", "NA", 0.7)
        self.cmf = _readconfig(config.get, "viewer", "cmf", "CIE1931")
        
        self.diffraction = _readconfig(config.getint, "transfer", "diffraction", 1)
        self.nin = _readconfig(config.getfloat, "transfer", "nin", self.n_cover)
        self.nout = _readconfig(config.getfloat, "transfer", "nout", self.n_cover)
        self.method = _readconfig(config.get, "transfer", "method", "2x2")
        self.npass = _readconfig(config.getint, "transfer", "npass", 1)
        self.reflection = _readconfig(config.getint, "transfer", "reflection", None)
        self.eff_data = _readconfig(config.getint, "transfer", "eff_data", 0)
        
    def __getitem__(self, item):
        return self.__dict__[item]
        
    def __repr__(self):
        return repr(self.__dict__)
    


#: a singleton holding user configuration    
DTMMConfig = DTMMConfig()
if DTMMConfig.nthreads > 1:
    disable_mkl_threading()

    
def get_default_config_option(name, value = None):
    """Returns default config option specified with 'name', if value is not None,
    returns value instead"""
    return DTMMConfig[name] if value is None else value
    
CMF = _readconfig(config.get, "viewer", "cmf", "CIE1931")

def print_config():
    """Prints all compile-time and run-time configurtion parameters and settings."""
    options = {"PRECISION" : PRECISION, "BETAMAX": BETAMAX, "SMOOTH" : SMOOTH,
               "NUMBA_FASTMATH" :NUMBA_FASTMATH, "NUMBA_PARALLEL" : NUMBA_PARALLEL,
           "NUMBA_CACHE" : NUMBA_CACHE, "NUMBA_FASTMATH" : NUMBA_FASTMATH,
           "NUMBA_TARGET" : NUMBA_TARGET}
    options.update(DTMMConfig.__dict__)
    print(options)


#setter functions for DTMMConfig    
def set_verbose(level):
    """Sets verbose level (0-2) used by compute functions."""
    out = DTMMConfig.verbose
    DTMMConfig.verbose = max(0,int(level))
    return out
    
def set_nthreads(num):
    """Sets number of threads used by fft functions."""
    out = DTMMConfig.nthreads
    DTMMConfig.nthreads = max(1,int(num))
    return out
   
def set_cache(level):
    """Sets compute cache level."""
    out = DTMMConfig.cache
    level = max(int(level),0)
    if level > 1:
        warnings.warn("Cache levels higher than 1 not supported yet!")
    DTMMConfig.cache = level
    return out

def set_fftlib(name = "numpy.fft"):
    """Sets fft library. Returns previous setting."""
    out, name = DTMMConfig.fftlib, str(name) 
    if name == "mkl_fft":
        if MKL_FFT_INSTALLED: 
            DTMMConfig.fftlib = name
        else:
            warnings.warn("MKL FFT is not installed so it can not be used! Please install mkl_fft.")            
    elif name == "scipy.fftpack" or name == "scipy":
        if SCIPY_INSTALLED:
            DTMMConfig.fftlib = "scipy"
        else:
            warnings.warn("Scipy is not installed so it can not be used! Please install scipy.") 
    elif name == "numpy.fft" or name == "numpy":
        DTMMConfig.fftlib = "numpy"
    else:
        raise ValueError("Unsupported fft library!")
    return out    

set_fftlib(_readconfig(config.get, "fft", "fftlib", DTMMConfig.fftlib))

import numba

NF32DTYPE = numba.float32
NF64DTYPE = numba.float64
NC64DTYPE = numba.complex64
NC128DTYPE = numba.complex128
NU32DTYPE = numba.uint32

if read_environ_variable("DTMM_DOUBLE_PRECISION",
  default =  _readconfig(config.getboolean, "core", "double_precision", True)):
    NFDTYPE = NF64DTYPE
    NCDTYPE = NC128DTYPE
    NUDTYPE = NU32DTYPE
else:
    NFDTYPE = NF32DTYPE
    NCDTYPE = NC64DTYPE
    NUDTYPE = NU32DTYPE
  
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
FFT mode selection
"""
import numpy as np
from dtmm.wave import betaxy
from dtmm.conf import cached_result, BETAMAX, FDTYPE, NFDTYPE, NUMBA_CACHE,NUMBA_TARGET
import numba as nb

@nb.guvectorize([(NFDTYPE[:,:],NFDTYPE[:,:],NFDTYPE[:],NFDTYPE[:],NFDTYPE[:],NFDTYPE[:],NFDTYPE[:],NFDTYPE[:],NFDTYPE[:],NFDTYPE[:,:])], 
                 "(i,j),(i,j),(),(),(),(),(),(),()->(i,j)",
                 target = NUMBA_TARGET, cache = NUMBA_CACHE)
def fft_window(fftbetax, fftbetay,betax, betay, stepx,stepy, xtyp, ytyp, betamax, out):
    ni, nj = fftbetax.shape
    for i in range(ni):
        for j in range(nj):
            fftbetaxij = fftbetax[i,j]
            dx = (fftbetaxij - betax[0])/stepx[0]
            cx = 1. - np.abs(dx)
            if cx < 0.:
                cx = 0.  
            else:
                if xtyp[0] < 0:
                    if dx < 0:
                        cx = 1.
                elif xtyp[0] > 0:
                    if dx > 0:
                        cx = 1.
            fftbetayij = fftbetay[i,j]
            dy = (fftbetayij - betay[0])/stepy[0]
            cy = 1. - np.abs(dy)
            if cy < 0.:
                cy = 0.
            else:
                if ytyp[0] < 0:
                    if dy < 0:
                        cy = 1.
                elif ytyp[0] > 0:
                    if dy > 0:
                        cy = 1.
            beta = (fftbetaxij**2 + fftbetayij**2)**0.5 
            if beta >= betamax[0]:
                coeff = 0.
            else:
                coeff = cx * cy
            if coeff > 0.:
                out[i,j] = coeff
            else:
                out[i,j] = 0.

def fft_windows(betax, betay, n, betax_off = 0., betay_off = 0., betamax = BETAMAX, out = None):

    d = 2*betamax/(n-1)
    xoffset = np.mod(betax_off, d)
    xoffsetm = np.mod(betax_off, -d) 
    mask = np.abs(xoffset) > np.abs(xoffsetm)
    try:
        xoffset[mask] = xoffsetm[mask]
    except TypeError: #scalar
        if mask:
            xoffset = xoffsetm
    
    yoffset = np.mod(betay_off, d)
    yoffsetm = np.mod(betay_off, -d) 
    mask = np.abs(yoffset) > np.abs(yoffsetm)
    try:
        yoffset[mask] = yoffsetm[mask]
    except TypeError:
        if mask:
            yoffset = yoffsetm
    
    ax = np.linspace(-betamax, betamax, n)
    ay = np.linspace(-betamax, betamax, n)
    step = ax[1]-ax[0]
    for i,bx in enumerate(ax):
        if i == 0:
            xtyp = -1
        elif i == n-1:
            xtyp = 1
        else:
            xtyp = 0
        for j,by in enumerate(ay):
            index = i*n+j
            if out is None:
                _out = None
            else:
                _out = out[index]
            
            if j == 0:
                ytyp = -1
            elif j == n-1:
                ytyp = 1
            else:
                ytyp = 0
            
            fmask = fft_window(betax,betay,bx+xoffset,by+yoffset,step,step,xtyp,ytyp,betamax, out = _out) 
            if out is None:
                out = np.empty((n*n,)+fmask.shape, fmask.dtype)
                out[0] = fmask
        
    return out

def fft_betaxy(shape, k0):
    bx,by = betaxy(shape[-2:], np.asarray(k0,FDTYPE)[...,None])
    return bx,by #np.broadcast_to(bx,shape),np.broadcast_to(by,shape)

def fft_betaxy_mean(betax, betay, fft_windows):
    axis = tuple(range(-betax.ndim,0))
    bx = betax*fft_windows
    by = betay*fft_windows
    norm = fft_windows.sum(axis = axis)
    bx = bx.sum(axis = axis) 
    by = by.sum(axis = axis) 
    mask = (norm>0)
    return np.divide(bx,norm, where = mask, out = bx), np.divide(by,norm, where = mask, out = by)

def betaxy2betaphi(bx,by):
    beta = (bx**2+by**2)**0.5
    phi = np.arctan2(by,bx)
    return beta, phi

@cached_result
def fft_mask_full(shape, k0, n, betax_off = 0., betay_off = 0., betamax = BETAMAX):
    
    betax, betay = fft_betaxy(shape, k0)
    windows = fft_windows(betax, betay, n, betax_off = betax_off, betay_off = betay_off, betamax = betamax)
    bxm, bym = fft_betaxy_mean(betax, betay, windows)

    return windows, betaxy2betaphi(bxm,bym)

@cached_result
def fft_mask(shape, k0, n, betax_off = 0., betay_off = 0., betamax = BETAMAX):
    windows, (bs,ps) = fft_mask_full(shape, k0, n, betax_off, betay_off, betamax)
    zero = np.asarray([np.alltrue(w == 0.) for w in windows])
    nonzero = np.logical_not(zero)
    
    return windows[nonzero], (bs[nonzero], ps[nonzero])
    
    
if __name__ == "__main__":
    w,bp = fft_mask_full((64,1), (1,), 3, betax_off = 0.)
    import matplotlib.pyplot as plt
    fig,axes = plt.subplots(3,3)
    for i in range(3):
        for j in range(3):
            n = i + 3*j
            axes[i,j].imshow(np.fft.fftshift(w[n,0,0]))
            axes[i,j].axis('off')

    plt.show()
    
    